# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""This module provides a container class for multidimensional scalar or vector fields."""


import logging
from numbers import Number
from numpy.lib.mixins import NDArrayOperatorsMixin

import numpy as np
from numpy.core import numeric
from scipy.ndimage import interpolation


__all__ = ['Field']


class Field(NDArrayOperatorsMixin):
    """Container class for storing multidimensional scalar or vector fields.

    The `Field` class is a sophisticated wrapper around a multidimensional numpy array. The user can access the
    underlying data via the `data` attribute, or by using the `numpy.asarray` function.

    `Field` defines the `ufunc` interface for numpys vast library of universal functions. This includes all operator
    definitions, i.e., you can add, subtract, multiply and devide `Field` objects element-wise. Note that for vector
    fields, `Field` will handle all vector components separately and for ufuncs with two inputs, the other object will
    be broadcast accordingly. It is therefore also possible to add (or multiply) vector and scalar fields, assuming
    their dimensions `dim` match (their `shape` can't match, because it includes the number of components `ncomp` for
    the vector field). For ufuncs that reduce the output shape (e.g. `numpy.sum`), if the `axis` parameter is set to
    `None` (default), component axes for vector fields are kept (they can however explicitely reduced by including
    them in the `axis` paremeter, e.g. by setting `axis = -1`).

    `Field` objects can be indexed like numpy arrays and scalar fields (`vector=False`) behave as expected. A vector
    field (`vector=True`) can be indexed by only the non-component indices (see `dim`) and the according components of
    the specified point are returned as a tuple. Slicing with integers will return a `Field` object with reduced
    dimensionality (or a scalar/tuple if all spatial dimensions are reduced) that also drops the associated `scales`
    as needed. New axes (indexing with `None` or `numpy.newaxis`) is disabled because that would mean that an unknown
    scale would need to be added (you can always create a new `Field` object with an appropriately prepared ndarray).

    Some callables exposed by numpy's public API (e.g. `numpy.mean(...)`) will treat `Field` objects as if they
    were identical to their underlying numpy array (by envoking the `__array__` function which just returns `data`).
    The result often is a simple `numpy.ndarray` and not a `Field` object. Some functions (e.g. `clip`) are also defined
    on `Field` as methods. They might behave differently and will return `Field` objects instead, so take care and
    refer to their docstrings for further information!
    # TODO: Too verbose? In documentation instead?

    Attributes
    ----------
    data: np.ndarray
        The underlying data array of the field.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.
    vector: bool
        True if the field is a vector field, False if it is a scalar field.
    shape: tuple of ints
        Shape of the underlying data. Includes the number of components as the first dimension if the field is a
        vector field.
    dim: tuple of ints
        Dimensions of the underlying data. Only includes spatial dimensions, NOT the number of vector components!
        (Use `ncomp` to get the number of components, if any, `shape` for the full shape of the underlying array).
    ncomp: None or tuple of ints
        Number of components if the field is a vector field (can be 2 or 3), None if it is a scalar field. The component
        axis is always the last axis (index `-1`) of the underlying data array!

    Notes
    -----
    See https://docs.scipy.org/doc/numpy/reference/arrays.classes.html for information about ndarray subclassing!
    See https://docs.scipy.org/doc/numpy-1.13.0/neps/ufunc-overrides.html for information about __array_ufunc__!
    See https://numpy.org/neps/nep-0018-array-function-protocol.html for information about __array_function__!
    """

    _log = logging.getLogger(__name__ + '.Field')

    @property
    def data(self):
        return self.__data

    @property
    def vector(self):
        return self.__vector

    @vector.setter
    def vector(self, vector):
        assert isinstance(vector, bool), 'vector has to be a boolean!'
        self.__vector = vector

    @property
    def scale(self):
        return self.__scale

    @scale.setter
    def scale(self, scale):
        if isinstance(scale, Number):  # Scale is the same for each dimension!
            self.__scale = (scale,) * len(self.dim)
        elif isinstance(scale, tuple):
            assert len(scale) == len(self.dim), f'Each dimension {self.dim} needs a scale, but {scale} was given!'
            self.__scale = scale
        else:
            raise AssertionError('Scaling has to be a number or a tuple of numbers!')

    @property
    def shape(self):
        return self.data.shape

    @property
    def dim(self):
        if self.vector:
            return self.shape[:-1]
        else:
            return self.shape

    @property
    def ncomp(self):
        return self.shape[-1] if self.vector else 0

    @property
    def comp(self):  # TODO: Function or property? Philosophical question?
        # Return empty list for scalars (ncomp is 0) and a list of components as scalar fields for a vector field:
        return [Field(self.data[..., i], self.scale, vector=False) for i in range(self.ncomp)]

    @property
    def amp(self):  # TODO: Function or property? Philosophical question?
        if self.vector:
            amp = np.sqrt(np.sum([comp.data**2 for comp in self.comp], axis=0))
        else:
            amp = np.abs(self.data)
        return Field(amp, self.scale, vector=False)

    @property
    def mask(self):  # TODO: Function or property? Philosophical question?
        return Field(np.where(np.asarray(self.amp) > 0, True, False), self.scale, vector=False)

    def __init__(self, data, scale=1, vector=False):
        self._log.debug('Calling __init__')
        self.__data = data
        self.vector = vector  # Set vector before scale, because scale needs to know if vector (via calling dim)!
        self.scale = scale
        if self.vector:
            assert self.ncomp in (2, 3), 'Only 2- or 3-dimensional vector fields are supported!'
        self._log.debug('Created ' + str(self))

    def __repr__(self):
        self._log.debug('Calling __repr__')
        data_string = str(self.data)  # String of JUST the data array, without metadata (compared to repr)!
        return f'{self.__class__.__name__}(data={data_string}, scale={self.scale}, vector={self.vector})'

    def __str__(self):
        self._log.debug('Calling __repr__')
        ncomp_string = f', ncomp={self.ncomp}' if self.vector else ''
        return f'{self.__class__.__name__}(dim={self.dim}, scale={self.scale}, vector={self.vector}{ncomp_string})'

    def __array__(self):
        self._log.debug('Calling __array__')
        return self.data

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        self._log.debug('Calling __array_ufunc__')
        outputs = kwargs.pop('out', ())
        outputs = kwargs.pop('out', (None,)*ufunc.nout)  # Defaults to tuple of None (currently: nout=1 all the time)
        outputs_arr = tuple([np.asarray(out) if isinstance(out, Field) else out for out in outputs])
        # Cannot handle items that have __array_ufunc__ (other than our own).
        for item in inputs + outputs:
            if hasattr(item, '__array_ufunc__') and not isinstance(item, Field):  # Something else with __array_ufunc__:
                if type(item).__array_ufunc__ is not np.ndarray.__array_ufunc__:  # Can't handle other overrides:
                    return NotImplemented
        # TODO: BIGGEST NOTE HERE: Delegate work to ndarray.__array_ufunc__!
        # TODO: For this to work, we have to make sure that we dispatch input as ndarrays, NOT Fields!
        # Convert all Field inputs to simple ndarrays to avoid infinite recursion!
        # ndarray.__array_ufunc__ delegates to the called ufunc method ONLY if all inputs and outputs are ndarrays
        # (or have no __array_ufunc__ method, e.g. a scalar), so that's what we need to ensure here.:
        # inputs = tuple([inp.view(np.ndarray) if isinstance(inp, Field) else inp for inp in inputs])
        # TODO: possibly need to sort out which input (if two) is a Field (vector or scalar?), order matters?
        # TODO: DOCUMENT: most things are same as in ndarrays, but if only one input is a vector field, try broadcast!
        # TODO: for security, newaxis is not allowed (most other indexing works though), because scale would be unknown!
        # 1 input (has to be a Field, otherwise we wouldn't be here):
        if len(inputs) == 1:
            self._log.debug(f'__array_ufunc__ inputs: {len(inputs)}')
            field = inputs[0]
            scale_new = field.scale
            vector_new = field.vector
            # Preprocess axis keyword if it exists:
            axis = kwargs.get('axis', False)  # Default must not be None, because None is a possible setting!
            full_reduction = False
            if axis is not False:
                ax_full = tuple(range(len(field.dim)))  # All axes (minus a possible component axis for vector Fields)!
                ax_full_wc = tuple(range(len(field.dim) + 1))  # All axes WITH component axis (does not need to exist)!
                axis = ax_full if axis is None else axis  # This keeps possible components untouched if axis was None!
                axis = numeric.normalize_axis_tuple(axis, field.ndim)  # Takes care of pot. neg. indices, ensures tuple!
                kwargs['axis'] = axis  # Put normalized axis back into kwargs!
                if axis in (ax_full, ax_full_wc):  # Full reduction (or reduction to just components) takes place:
                    full_reduction = True
                if ax_full_wc[-1] in axis:  # User explicitely wants component reduction (can only be true for vector):
                    vector_new = False  # Force scalar field!
                scale_new = tuple([s for i, s in enumerate(field.scale) if i not in axis])  # Drop axis from scale!
            inputs_arr = np.asarray(field)  # Convert inputs that are Fields to ndarrays to avoid recursion!
            data_new = getattr(ufunc, method)(inputs_arr, out=outputs_arr, **kwargs)
            if full_reduction:  # Premature return because the result is no longer a Field:
                return data_new
        # More than 1 input (at least one has to be a Field, otherwise we wouldn't be here):
        elif len(inputs) > 1:
            is_field = [isinstance(inp, Field) for inp in inputs]
            is_vector = [getattr(inp, 'vector', False) for inp in inputs]
            # Determine scale:
            if np.sum(is_field) > 1:  # More than one input is a Field objects:
                scales = [inp.scale for i, inp in enumerate(inputs) if is_field[i]]  # Only takes scales of Field obj.!
                scale_new = scales[0]
                err_msg = f'Scales of all Field objects must match! Given scales: {scales}!'
                assert all(scale == scale_new for scale in scales), err_msg
            else:  # Only one input is a field, pick the scale of that one:
                scale_new = inputs[np.argmax(is_field)].scale  # argmax returns the index of first True!
            # Determine vector:
            vector_new = True if np.any(is_vector) else False  # Output is vector field if any input is a vector field!
            if np.sum(is_vector) > 1:  # More than one input is a vector Field objects:
                ncomps = [inp.ncomp for i, inp in enumerate(inputs) if is_vector[i]]  # Only takes ncomp of v.-Fields!
                err_msg = f'# of components of all Field objects must match! Given ncomps: {ncomps}!'
                assert all(ncomp == ncomps[0] for ncomp in ncomps), err_msg
            # Append new axis at the end of non vector objects to broadcast to components:
            if np.any(is_vector):
                inputs = list(inputs)
                for i, inp in enumerate(inputs):
                    if not is_vector[i] and not isinstance(inp, Number):  # Numbers work for broadcasting anyway:
                        if len(np.asarray(inp).shape) == len(scale_new):  # No. of dimensions w/o comp., have to match!
                            inputs[i] = np.asarray(inputs[i])[..., np.newaxis]  # Broadcasting, try to cast as ndarray!
                inputs = tuple(inputs)
            # Convert inputs that are Fields to ndarrays to avoid recursion and determine data_new:
            inputs_arr = tuple([np.asarray(inp) if isinstance(inp, Field) else inp for inp in inputs])
            data_new = getattr(ufunc, method)(*inputs_arr, out=outputs_arr, **kwargs)
        # Return results:
        result = Field(data_new, scale_new, vector_new)
        return result

    def __getitem__(self, index):
        self._log.debug('Calling __getitem__')
        # Pre-processing index:
        index = index if isinstance(index, tuple) else (index,)  # Make sure item is a tuple!
        if None in index:
            raise Exception('Field does not support indexing with newaxis/None! Please cast as ndarray first!')
        if len(index) < len(self.shape) and Ellipsis not in index:  # Ellipsis is IMPLICIT at the end if index < dim:
            index += (Ellipsis,)
        if Ellipsis in index:  # Expand Ellipsis (...) into slice(None) (:) to make iterating consistent:
            index = list(index)
            i = index.index(Ellipsis)
            missing_dims = len(self.shape) - len(index)  # Number of non-specified dimensions
            index[i:i+1] = [slice(None)] * (missing_dims + 1)  # +1 because at least Ellipsis is replaced!
            index = tuple(index)
        # Indexing with a scalar drops the dimension, indexing with slice always keeps the dimension:
        index_scale = index[:-1] if self.vector else index  # Disregard last index for vector fields (has no scale)!
        scale_new = ()
        for i, item in enumerate(index_scale):
            if isinstance(item, slice):  # Indexing with slice keeps the dimension and therefore the scale:
                scale_new += (self.scale[i],)
        # Get data with classic indexing from underlying data array:
        data_new = self.data[index]
        # Determine vector state of output:
        vector_new = self.vector
        if self.vector and isinstance(index[-1], Number):  # For vector fields if last index (component) is dropped:
            vector_new = False
        # Return:
        if not scale_new:  # scale_new=(), i.e. full reduction of all dimensions (not regarding possible vector comp.):
            if isinstance(data_new, Number):  # full reduction:
                return data_new
            else:  # Only important for vector fields:
                return tuple(data_new)  # return single vector as tuple
        else:  # Return Field object
            return Field(data_new, scale_new, vector=vector_new)

    @classmethod
    def from_scalar_fields(cls, scalar_list):
        """Create a vector `Field` object from a list of scalar `Fields`.

        Parameters
        ----------
        scalar_list : list
            List of scalar `Field` objects (`vector=False`).

        Returns
        -------
        vector_field: `Field`
            A vector `Field` object containing the input scalar fields as components.

        """
        cls._log.debug('Calling from_scalar_fields')
        assert len(scalar_list) in (2, 3), 'Needs two or three scalar fields as components for vector field!'
        assert all(isinstance(item, Field) for item in scalar_list), 'All items have to be Field objects!'
        assert all(not item.vector for item in scalar_list), 'All items have to be scalar fields!'
        assert all(item.scale == scalar_list[0].scale for item in scalar_list), 'Scales of fields must match!'
        return cls(np.stack(scalar_list, axis=-1), scalar_list[0].scale, vector=True)

    def squeeze(self):
        """Squeeze the `Field` object to remove single-dimensional entries in the shape.

        Returns
        -------
        field: `Field`
            Squeezed `Field` object. Note that scales associated with squeezed dimensions are also dropped.

        Notes
        -----
        Also works via `numpy.squeeze(field)`, because `numpy.squeeze` searches for a local implementation first and
        then uses `_wrapit` to envoke this function here!

        """
        self._log.debug('Calling squeeze')
        squeezed_indices = np.flatnonzero(np.asarray(self.dim) == 1)
        squeezed_data = np.squeeze(self.data)
        if squeezed_indices:
            self._log.info(f'The following indices were squeezed: {squeezed_indices}')
        squeezed_scale = tuple([self.scale[i] for i in range(len(self.dim)) if i not in squeezed_indices])
        return Field(squeezed_data, squeezed_scale, self.vector)

    def pad(self, pad_width, mode='constant', **kwargs):
        """Pad the `Field` object and increase the size of the underlying array.

        Parameters
        ----------
        pad_width : {sequence, array_like, int}
            Number of values padded to the edges of each axis. Can be a single number for all, one number per axis or
            a tuple `(before, after)` for each axis for finer control.
        mode : str, optional
            Padding mode (see `numpy.pad` for all options), by default 'constant', which pads with zeros.

        Returns
        -------
        field: `Field`
            The padded `Field` object.

        """
        if isinstance(pad_width, Number):  # Paddding is the same for each dimension (make sure it is a tuple)!
            pad_width = (pad_width,) * len(self.dim)
        pad_width = [(p, p) if isinstance(p, Number) else p for p in pad_width]
        if self.vector:  # Append zeros to padding, so component axis stays as is:
            pad_width = pad_width + [(0, 0)]
        data_new = np.pad(self.data, pad_width, mode, **kwargs)
        return Field(data_new, self.scale, self.vector)

    def bin(self, n):
        """Bins data of the `Field` to decrease the size of the underlying array by averaging over a number of values.

        Parameters
        ----------
        n : {sequence, array_like, int}
            Number of entries along each axis over which the average is taken. Can be a single integer for all axes or a
            list, specifying the number of entries over which to average for each individual axis.

        Returns
        -------
        field: `Field`
            The binned `Field` object.

        Notes
        -----
        Padding takes place before binning to ensure the dimensions are multiples of `n`. The padding mode is `edge`.

        """
        self._log.debug('Calling bin')
        assert isinstance(n, (int, tuple)), 'n must be a positive integer or a tuple of positive integers!'
        if isinstance(n, Number):  # Binning is the same for each dimension (make sure it is a tuple)!
            n = (n,) * len(self.dim)
        assert all([n_i >= 0 for n_i in n]), 'All entries of n must be positive integers!'
        # Pad if necessary (use padded 'field' from here on), formula for clarity: (n - dim % n) % n
        pad_width = [(0, (n[i] - self.dim[i] % n[i]) % n[i]) for i in range(len(self.dim))]
        field = self.pad(pad_width, mode='edge')
        # Create new shape used for binning (mean over every second axis will be taken):
        bin_shape = list(np.ravel([(field.dim[i]//n[i], n[i]) for i in range(len(field.dim))]))
        mean_axes = np.arange(1, 2*len(field.dim), 2)  # every 2nd axis!
        if self.vector:  # Vector field:
            bin_shape += [field.ncomp]  # Append component axis (they stay unchanged)
        # Bin data and scale accordingly:
        data_new = field.data.reshape(bin_shape).mean(axis=tuple(mean_axes))
        scale_new = tuple([field.scale[i] * n[i] for i in range(len(field.dim))])
        return Field(data_new, scale_new, self.vector)

    def zoom(self, zoom, **kwargs):
        """Wrapper for the `scipy.ndimage.zoom` function.

        Parameters
        ----------
        zoom : float or sequence
            Zoom factor along the axes. If a float, `zoom` is the same for each axis. If a sequence, `zoom` needs to
            contain one value for each axis.

        Returns
        -------
        field: `Field`
            The zoomed in `Field` object.

        Notes
        -----
        As in `scipy.ndimage.zoom`, a spline order can be specified, which defaults to 3.

        """
        self._log.debug('Calling zoom')
        if not self.vector:  # Scalar field:
            data_new = interpolation.zoom(self.data, zoom, **kwargs)
        else:  # Vector field:
            comps = [np.asarray(comp) for comp in self.comp]
            data_new = np.stack([interpolation.zoom(comp, zoom, **kwargs) for comp in comps], axis=-1)
        if isinstance(zoom, Number):  # Zoom is the same for each dimension!
            zoom = (zoom,) * len(self.dim)
        scale_new = tuple([self.scale[i]/z for i, z in enumerate(zoom)])
        return Field(data_new, scale_new, self.vector)

    def gradient(self):
        """Returns the gradient of an N-dimensional scalar `Field`. Wrapper around the according numpy function.

        Returns
        -------
        gradients: ndarray or list of ndarray
            A set of ndarrays (or a single ndarray for 1D input), corresponding to the derivatives of the field with
            respect to each dimension.

        Notes
        -----
        The field is implicitely squeezed before the gradient is calculated!

        """
        self._log.debug('Calling gradient')
        assert not self.vector, 'Gradient can only be computed from scalar fields!'
        squeezed_field = self.squeeze()  # Gradient along dimension of length 1 does not work!
        gradients = np.gradient(np.asarray(squeezed_field), *squeezed_field.scale)
        if len(squeezed_field.dim) == 1:  # Only one gradient!
            return Field(gradients, squeezed_field.scale, vector=False)
        else:  # Multidimensional gradient (flip component order with [::-1], so that e.g x/y/z instead of z/y/x):
            return Field(np.stack(gradients[::-1], axis=-1), squeezed_field.scale, vector=True)

    def curl(self):
        """Returns the curl of an N-dimensional `Field`.

        Returns
        -------
        field: `Field`
            The curl/rotation.

        Notes
        -----
        The calculation depends on the input:
        3 dimensions, 3 components: Calculates the full 3D rotational vector field!
        2 dimensions, 2 components: Calculates the out-of-plane component of the curl as a 2D scalar field!
        2 dimensions, scalar: Calculates the planar rotatio as a 2D vector field!

        """
        self._log.debug('Calling curl')
        squeezed_field = self.squeeze()
        if len(squeezed_field.dim) == 3:  # 3 dimensions:
            if squeezed_field.ncomp == 3:  # 3 component vector field (standard case):
                self._log.debug('input: 3 dimensions, 3 components!')
                field_x, field_y, field_z = squeezed_field.comp
                grad_xx, grad_xy, grad_xz = field_x.gradient().comp
                grad_yx, grad_yy, grad_yz = field_y.gradient().comp
                grad_zx, grad_zy, grad_zz = field_z.gradient().comp
                curl_x = grad_zy - grad_yz
                curl_y = grad_xz - grad_zx
                curl_z = grad_yx - grad_xy
                return Field.from_scalar_fields([curl_x, curl_y, curl_z])
            else:
                raise AssertionError('Can only handle 3 component vector fields in 3D!')
        elif len(squeezed_field.dim) == 2:  # 2 dimensions (tricky, usually not hardly defined):
            if squeezed_field.ncomp == 2:  # 2 component vector field (return perpendicular component as scalar field):
                self._log.debug('input: 2 dimensions, 2 components!')
                field_x, field_y = squeezed_field.comp
                grad_xx, grad_xy = field_x.gradient().comp
                grad_yx, grad_yy = field_y.gradient().comp
                return grad_yx - grad_xy
            elif not squeezed_field.vector:  # scalar field (return planar components as 2D vector field):
                self._log.debug('input: 3 dimensions, scalar field!')
                grad_x, grad_y = squeezed_field.gradient().comp
                curl_x = grad_y
                curl_y = -grad_x
                return Field.from_scalar_fields([curl_x, curl_y])
            else:
                raise AssertionError('Can only handle 3 component vector or scalar fields in 2D!')
        else:
            raise AssertionError('Can only handle 3D or 2D cases (see documentation for specifics)!')

    def clip(self, vmin=None, vmax=None, sigma=None, mask=None):
        """Clip (limit) the values in an array. For vector fields, this will take the amplitude into account

        Parameters
        ----------
        vmin : float, optional
            Mimimum value, by default None. Is ignored for vector fields. Will overwrite values found via `sigma` or
            `mask`.
        vmax : float, optional
            Maximum value, by default None. Will overwrite values found via `sigma` or `mask`.
        sigma : float, optional
            Defines a range in standard deviations, that results in a boolean mask that excludes outliers, by default
            None. E.g. `sigma=2` will mark points within the 5% highest amplitude as outliers. `vmin` and `vmax` will be
            searched in the remaining region. Is logically combined with `mask` if both are set.
        mask : ndarray, optional
            Boolean array that directly describes where to search for `vmin` and `vmax`, by default None. Is logically
            combined with the mask from `sigma` if both are set.

        Returns
        -------
        field: `Field`
            The clipped `Field`.

        Notes
        -----
        In contrast to the corresponding numpy function, `vmin` and `vmax` can both be `None` due to the alternative
        clipping strategies employed here.

        """
        self._log.debug('Calling clip')
        amp = self.amp.data
        if mask is None:  # If no mask is set yet, default to True everywhere:
            mask = np.full(self.dim, True)
        if sigma is not None:  # Mark outliers that are outside `sigma` standard deviations:
            sigma_mask = (amp - amp.mean()) < (sigma * amp.std())
            mask = np.logical_and(mask, sigma_mask)
        # Determine vmin and vmax if they are not set by the user:
        amp_masked = np.where(mask, amp, np.nan)
        if vmin is None:
            vmin = np.nanmin(amp_masked)
        if vmax is None:
            vmax = np.nanmax(amp_masked)
        if self.vector:  # Vector fields need to scale components according to masked amplitude
            mask_vec = np.logical_and(mask, amp <= vmax)  # Only vmax is important!
            data = amp / np.where(mask_vec, 1, amp)  # TODO: needs testing!
            # TODO: Test np.clip(field) (also ufunc?? think not...) seems to work, but what about others?
        else:  # For scalar fields, just delegate to the numpy function:
            data = np.clip(self.data, vmin, vmax)
        return Field(data, self.scale, self.vector)



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""Subpackage containing functionality for visualisation of multidimensional fields."""

from . import colors
from .plot2d import *
from .decorators import *
from .tools import *


__all__ = ['colors']
__all__.extend(plot2d.__all__)
__all__.extend(decorators.__all__)
__all__.extend(tools.__all__)


del decorators
del plot2d
del tools



# -*- coding: utf-8 -*-
# Copyright 2016 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#


import logging
from numbers import Number

import numpy as np

from .field import Field


__all__ = ['create_vector_homog', 'create_vector_vortex', 'create_vector_skyrmion', 'create_vector_singularity']

_log = logging.getLogger(__name__)


def create_vector_homog(dim, phi=0, theta=None, scale=1):
    """Field subclass implementing a homogeneous vector field with 3 components in 2 or 3 dimensions.

    Attributes
    ----------
    dim : tuple
        The dimensions of the grid.
    phi : float
        The azimuthal angle. The default is 0, i.e., the vectors point in x direction.
    theta : float, optional
        The polar angle. If None (default), only two components will be created (corresponds to pi/2 in 3D, i.e., the
        vectors are in the xy plane).
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    assert len(dim) in (2, 3), 'Disc can only be build in 2 or 3 dimensions!'
    assert isinstance(phi, Number), 'phi has to be an angle in radians!'
    assert isinstance(theta, Number) or theta is None, 'theta has to be an angle in radians or None!'
    if theta is None:
        y_comp = np.ones(dim) * np.sin(phi)
        x_comp = np.ones(dim) * np.cos(phi)
        data = np.stack([x_comp, y_comp], axis=-1)
    else:
        z_comp = np.ones(dim) * np.cos(theta)
        y_comp = np.ones(dim) * np.sin(theta) * np.sin(phi)
        x_comp = np.ones(dim) * np.sin(theta) * np.cos(phi)
        data = np.stack([x_comp, y_comp, z_comp], axis=-1)
    return Field(data=data, scale=scale, vector=True)


def create_vector_vortex(dim, center=None, core_r=0, oop_r=None, axis=0, scale=1):
    # TODO: oop and core_r documentation!!! General description!
    # TODO: CHIRALITY
    """Field subclass implementing a vortex vector field with 3 components in 2 or 3 dimensions.

    Attributes
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple (N=2 or N=3), optional
        The vortex center, given in 2D `(v, u)` or 3D `(z, y, x)`, where the perpendicular axis is discarded
        (determined by the `axis` parameter). Is set to the center of the field of view if not specified.
        The vortex center should be between two pixels to avoid singularities.
    axis :  int, optional
        The orientation of the vortex axis. The default is 0 and corresponds to the z-axis. Is ignored if dim is
        only 2D. To invert chirality, multiply the resulting Field object by -1.  # TODO: WRONG?

    """
    _log.debug('Calling create_vector_vortex')
    assert len(dim) in (2, 3), 'Vortex can only be build in 2 or 3 dimensions!'
    # Find indices of the vortex plane axes:
    idx_uv = [0, 1, 2]
    if len(dim) == 3:  # 3D:
        idx_uv.remove(axis)
    else:  # 2D:
        idx_uv.remove(2)
    # 2D dimensions:
    dim_uv = tuple([dim[i] for i in idx_uv])
    # Find default values:
    if center is None:
        center = tuple([dim[i] / 2 for i in idx_uv])
    elif len(center) == 3:  # if a 3D-center is given, just take the relevant coordinates:
        center = list(center)
        del center[axis]
        center = tuple(center)
    # Create vortex plane (2D):
    coords_uv = np.indices(dim_uv) + 0.5  # 0.5 to get to pixel/voxel center!
    coords_uv = coords_uv - np.asarray(center, dtype=float)[:, None, None]  # Shift by center!
    phi = np.arctan2(coords_uv[0], coords_uv[1]) - np.pi / 2
    rr = np.hypot(coords_uv[0], coords_uv[1])
    rr_clip = np.clip(rr - core_r, a_min=0, a_max=None)  # negative inside core_r (clipped to 0), positive outside!
    if oop_r is None:
        w_comp = np.zeros(dim_uv)
    else:
        w_comp = 1 - 2/np.pi * np.arcsin(np.tanh(np.pi*rr_clip/oop_r))  # orthogonal: 1 inside, towards 0 outside!
    v_comp = np.ones(dim_uv) * np.sin(phi) * np.sqrt(1 - w_comp)
    u_comp = np.ones(dim_uv) * np.cos(phi) * np.sqrt(1 - w_comp)
    if len(dim) == 3:  # Expand to 3D:
        w_comp = np.expand_dims(w_comp, axis=axis)
        v_comp = np.expand_dims(v_comp, axis=axis)
        u_comp = np.expand_dims(u_comp, axis=axis)
        reps = [1, 1, 1]  # repetitions for tiling
        reps[axis] = dim[axis]  # repeat along chosen axis
        w_comp = np.tile(w_comp, reps)
        v_comp = np.tile(v_comp, reps)
        u_comp = np.tile(u_comp, reps)
    if axis == 0:  # z-axis
        z_comp = w_comp
        y_comp = -v_comp
        x_comp = -u_comp
    elif axis == 1:  # y-axis
        z_comp = v_comp
        y_comp = w_comp
        x_comp = u_comp
    elif axis == 2:  # x-axis
        z_comp = -v_comp
        y_comp = -u_comp
        x_comp = w_comp
    else:
        raise ValueError(f'{axis} is not a valid argument for axis (has to be 0, 1 or 2)')
    data = np.stack([x_comp, y_comp, z_comp], axis=-1)
    return Field(data=data, scale=scale, vector=True)


def create_vector_skyrmion(dim, center=None, phi_0=0, skyrm_d=None, wall_d=None, axis=0, scale=1):
    """Create a 3-dimensional magnetic Bloch or Neel type skyrmion distribution.

    Parameters
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple (N=2 or N=3), optional
        The source center, given in 2D `(v, u)` or 3D `(z, y, x)`, where the perpendicular axis
        is discarded. Is set to the center of the field of view if not specified.
        The center has to be between two pixels.
    phi_0 : float, optional
        Angular offset switching between Neel type (0 [default] or pi) or Bloch type (+/- pi/2)
        skyrmions.
    skyrm_d : float, optional
        Diameter of the skyrmion. Defaults to half of the smaller dimension perpendicular to the
        skyrmion axis if not specified.
    wall_d : float, optional
        Diameter of the domain wall of the skyrmion. Defaults to `skyrm_d / 4` if not specified.
    axis :  {'z', '-z', 'y', '-y', 'x', '-x'}, optional # TODO: NUMBERS!! See vortex
        The orientation of the skyrmion axis. The default is 'z'. Negative values invert skyrmion
        core direction.

    Returns
    -------
    amplitude : tuple (N=3) of :class:`~numpy.ndarray` (N=3)
        The magnetic distribution as a tuple of the 3 components in
        `x`-, `y`- and `z`-direction on the 3-dimensional grid.

    Notes
    -----
        To avoid singularities, the source center should lie between the pixel centers (which
        reside at coordinates with _.5 at the end), i.e. integer values should be used as center
        coordinates (e.g. coordinate 1 lies between the first and the second pixel).

        Skyrmion wall width is dependant on exchange stiffness  A [J/m] and anisotropy K [J/m³]
        The out-of-plane magnetization at the domain wall can be described as:
        Mz = -Ms * tanh(x/w)  # TODO: Instead ROMER paper
        w = sqrt(A/K)

    """

    def _theta(r):
        theta_1 = + np.arcsin(np.tanh((r + skyrm_d/2)/(wall_d/2)))
        theta_2 = - np.arcsin(np.tanh((r - skyrm_d/2)/(wall_d/2)))
        theta = theta_1 + theta_2
        theta /= np.abs(theta).max() / np.pi
        return theta

    _log.debug('Calling create_vector_skyrmion')
    assert len(dim) in (2, 3), 'Skyrmion can only be build in 2 or 3 dimensions!'
    # Find indices of the skyrmion plane axes:
    idx_uv = [0, 1, 2]
    if len(dim) == 3:  # 3D:
        idx_uv.remove(axis)
    else:  # 2D:
        idx_uv.remove(2)
    # 2D dimensions:
    dim_uv = tuple([dim[i] for i in idx_uv])
    # Find default values:
    if skyrm_d is None:
        skyrm_d = np.min(dim_uv) / 2
    if wall_d is None:
        wall_d = skyrm_d / 4
    if center is None:
        center = tuple([dim[i] / 2 for i in idx_uv])
    elif len(center) == 3:  # if a 3D-center is given, just take the relevant coordinates:
        center = list(center)
        del center[axis]
        center = tuple(center)
    # Create skyrmion plane (2D):
    coords_uv = np.indices(dim_uv) + 0.5  # 0.5 to get to pixel/voxel center!
    coords_uv = coords_uv - np.asarray(center, dtype=float)[:, None, None]  # Shift by center!
    rr = np.hypot(coords_uv[0], coords_uv[1])
    phi = np.arctan2(coords_uv[0], coords_uv[1]) - phi_0
    theta = _theta(rr)
    w_comp = np.cos(theta)
    v_comp = np.sin(theta) * np.sin(phi)
    u_comp = np.sin(theta) * np.cos(phi)
    # Expansion to 3D if necessary and component shuffling:
    if len(dim) == 3:  # Expand to 3D:
        w_comp = np.expand_dims(w_comp, axis=axis)
        v_comp = np.expand_dims(v_comp, axis=axis)
        u_comp = np.expand_dims(u_comp, axis=axis)
        reps = [1, 1, 1]  # repetitions for tiling
        reps[axis] = dim[axis]  # repeat along chosen axis
        w_comp = np.tile(w_comp, reps)
        v_comp = np.tile(v_comp, reps)
        u_comp = np.tile(u_comp, reps)
    if axis == 0:  # z-axis
        z_comp = w_comp
        y_comp = -v_comp
        x_comp = -u_comp
    elif axis == 1:  # y-axis
        z_comp = v_comp
        y_comp = w_comp
        x_comp = u_comp
    elif axis == 2:  # x-axis
        z_comp = -v_comp
        y_comp = -u_comp
        x_comp = w_comp
    else:
        raise ValueError(f'{axis} is not a valid argument for axis (has to be 0, 1 or 2)')
    data = np.stack([x_comp, y_comp, z_comp], axis=-1)
    return Field(data=data, scale=scale, vector=True)


def create_vector_singularity(dim, center=None, scale=1):
    """Create a 3-dimensional magnetic distribution of a homogeneous magnetized object.

    Parameters
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple (N=2 or N=3), optional
        The source center, given in 2D `(v, u)` or 3D `(z, y, x)`, where the perpendicular axis
        is discarded. Is set to the center of the field of view if not specified.
        The source center has to be between two pixels.
    axis :  {'z', '-z', 'y', '-y', 'x', '-x'}, optional  # TODO: NUMBERS!
        The orientation of the source axis. The default is 'z'. Negative values invert the source
        to a sink.  # TODO: wat...?
    # TODO: scale!

    Returns
    -------
    amplitude : tuple (N=3) of :class:`~numpy.ndarray` (N=3)
        The magnetic distribution as a tuple of the 3 components in
        `x`-, `y`- and `z`-direction on the 3-dimensional grid.

    Notes
    -----
        To avoid singularities, the source center should lie between the pixel centers (which
        reside at coordinates with _.5 at the end), i.e. integer values should be used as center
        coordinates (e.g. coordinate 1 lies between the first and the second pixel).

    """  # TODO: What does negating do here? Senke / Quelle (YES IT DOES!)? swell and sink? ISSUE! INVITE PEOPLE!!!
    _log.debug('Calling create_vector_singularity')
    # Find default values:
    if center is None:
        center = tuple([d / 2 for d in dim])
    assert len(dim) == len(center), f"Length of dim ({len(dim)}) and center ({len(center)}) don't match!"
    # Setup coordinates, shape is (c, z, y, x), if 3D, or (c, y, x), if 2D (c: components):
    coords = np.indices(dim) + 0.5  # 0.5 to get to pixel/voxel center!
    center = np.asarray(center, dtype=float)
    bc_shape = (len(dim,),) + (1,)*len(dim)  # Shape for broadcasting, (1,1,1,3) for 3D, (1,1,2) for 2D!
    coords = coords - center.reshape(bc_shape)  # Shift by center (append 1s for broadcasting)!
    rr = np.sqrt(np.sum([coords[i]**2 for i in range(len(dim))], axis=0))
    data = coords / (rr + 1E-30)  # Normalise amplitude (keep direction), rr (z,y,x) is broadcasted to data (c,z,y,x)!
    data = data.T  # (c,z,y,x) -> (x,y,z,c)
    return Field(data=data, scale=scale, vector=True)



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#

"""This module provides a number of custom colormaps, which also have capabilities for 3D plotting.
If this is the case, the :class:`~.Colormap3D` colormap class is a parent class. In `cmaps`, a
number of specialised colormaps is available for convenience. If the default for circular colormaps
(used for 3D plotting) should be changed, set it via `CMAP_CIRCULAR_DEFAULT`.
For general questions about colors see:
http://www.poynton.com/PDFs/GammaFAQ.pdf
http://www.poynton.com/PDFs/ColorFAQ.pdf
"""

import logging
import colorsys
import abc

import numpy as np
from PIL import Image
from matplotlib import colors
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedLocator
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401
from matplotlib.patches import Circle

from .tools import use_style


__all__ = ['Colormap3D', 'ColormapCubehelix', 'ColormapPerception', 'ColormapHLS', 'ColormapClassic',
           'ColormapTransparent', 'cmaps', 'CMAP_CIRCULAR_DEFAULT', 'interpolate_color']

_log = logging.getLogger(__name__)


class Colormap3D(colors.Colormap, metaclass=abc.ABCMeta):
    """Colormap subclass for encoding directions with colors.

    This abstract class is used as a superclass/interface for 3D vector plotting capabilities.
    In general, a circular colormap should be used to encode the in-plane angle (hue). The
    perpendicular angle is encoded via luminance variation (up: white, down: black). Finally,
    the length of a vector is encoded via saturation. Decreasing vector length causes a desaturated
    color. Subclassing colormaps get access to routines to plot a colorwheel (which should
    ideally be located in the 50% luminance plane, which depends strongly on the underlying map),
    a convenience function to interpolate color tuples and a function to return rgb triples for a
    given vector. The :class:`~.Colormap3D` class itself subclasses the matplotlib base colormap.

    """

    _log = logging.getLogger(__name__ + '.Colormap3D')

    def rgb_from_vector(self, vector, vmax=None):
        """Construct a hls tuple from three coordinates representing a 3D direction.

        Parameters
        ----------
        vector: tuple (N=3) or :class:`~numpy.ndarray`
            Vector containing the x, y and z component, or a numpy array encompassing the
            components as three lists.

        Returns
        -------
        rgb:  :class:`~numpy.ndarray`
            Numpy array containing the calculated color tuples.

        """
        self._log.debug('Calling rgb_from_vector')
        x, y, z = np.asarray(vector)
        R = np.sqrt(x ** 2 + y ** 2 + z ** 2)
        R_max = vmax if vmax is not None else R.max() + 1E-30
        # FIRST color dimension: HUE (1D ring/angular direction)
        phi = np.asarray(np.arctan2(y, x))
        phi[phi < 0] += 2 * np.pi
        hue = phi / (2 * np.pi)
        rgba = np.asarray(self(hue))
        r, g, b = rgba[..., 0], rgba[..., 1], rgba[..., 2]
        # SECOND color dimension: SATURATION (2D, in-plane)
        rho = np.sqrt(x ** 2 + y ** 2)
        sat = rho / R_max
        r, g, b = interpolate_color(sat, (0.5, 0.5, 0.5), np.stack((r, g, b), axis=-1))
        # THIRD color dimension: LUMINANCE (3D, color sphere)
        theta = np.arccos(z / R_max)
        lum = 1 - theta / np.pi  # goes from 0 (black) over 0.5 (grey) to 1 (white)!
        lum_target = np.where(lum < 0.5, 0, 1)  # Separate upper(white)/lower(black) hemispheres!
        lum_target = np.stack([lum_target] * 3, axis=-1)  # [0, 0, 0] -> black / [1, 1, 1] -> white!
        fraction = 2 * np.abs(lum - 0.5)  # 0.5: difference from grey, 2: scale to range (0, 1)!
        r, g, b = interpolate_color(fraction, np.stack((r, g, b), axis=-1), lum_target)
        # Return RGB:
        return np.asarray(255 * np.stack((r, g, b), axis=-1), dtype=np.uint8)

    def make_colorwheel(self, size=64):
        """Construct a color wheel as an :class:`~PIL.Image` object.

        Parameters
        ----------
        size : int, optional
            Diameter of the color wheel along both axes in pixels, by default 64.

        Returns
        -------
        img: :class:`~PIL.Image`
            The resulting image.

        """
        self._log.debug('Calling make_colorwheel')
        # Construct the colorwheel:
        yy, xx = (np.indices((size, size)) - size/2 + 0.5)
        rr = np.hypot(xx, yy)
        xx = np.where(rr <= size/2-3, xx, 0)
        yy = np.where(rr <= size/2-3, yy, 0)
        zz = np.zeros((size, size))
        aa = np.where(rr >= size/2-3, 0, 255).astype(dtype=np.uint8)
        rgba = np.dstack((self.rgb_from_vector(np.asarray((xx, yy, zz))), aa))
        # Create color wheel:
        return Image.fromarray(rgba)

    def plot_colorwheel(self, axis=None, size=64, arrows=True, grayscale=False, **kwargs):
        """Display a color wheel to illustrate the color coding of vector gradient directions.

        Parameters
        ----------
        figsize : tuple of floats (N=2)
            Size of the plot figure.

        Returns
        -------
        img: :class:`matplotlib.image.AxesImage`
            The resulting colorwheel.

        """
        self._log.debug('Calling plot_colorwheel')
        # Construct the colorwheel:
        color_wheel = self.make_colorwheel(size=size)
        if grayscale:
            color_wheel = color_wheel.convert('LA')
        # Plot the color wheel:
        with use_style('empyre-image'):  # Only works on axes created WITHIN context!
            if axis is None:  # If no axis is set, find the current or create a new one:
                fig = plt.figure()
                axis = fig.add_subplot(1, 1, 1, aspect='equal')
            # Plot:
            im = axis.imshow(color_wheel, **kwargs)
            xy = size/2 - 0.5, size/2 - 0.5
            axis.add_patch(Circle(xy=xy, radius=size/2-2.5, linewidth=2, edgecolor='k', facecolor='none'))
            if arrows:
                axis.arrow(size/2, size/2+5, 0, 0.1*size, fc='k', ec='k', lw=1, width=2, alpha=0.15)
                axis.arrow(size/2, size/2-5, 0, -0.1*size, fc='k', ec='k', lw=1, width=2, alpha=0.15)
                axis.arrow(size/2+5, size/2, 0.1*size, 0, fc='k', ec='k', lw=1, width=2, alpha=0.15)
                axis.arrow(size/2-5, size/2, -0.1*size, 0, fc='k', ec='k', lw=1, width=2, alpha=0.15)
            return im


class ColormapCubehelix(colors.LinearSegmentedColormap, Colormap3D):
    """A full implementation of Dave Green's "cubehelix" for Matplotlib.

    Based on the FORTRAN 77 code provided in D.A. Green, 2011, BASI, 39, 289.
    http://adsabs.harvard.edu/abs/2011arXiv1108.5083G
    Also see:
    http://www.mrao.cam.ac.uk/~dag/CUBEHELIX/
    http://davidjohnstone.net/pages/cubehelix-gradient-picker
    User can adjust all parameters of the cubehelix algorithm. This enables much greater
    flexibility in choosing color maps. Default color map settings produce the standard cubehelix.
    Create color map in only blues by setting rot=0 and start=0. Create reverse (white to black)
    backwards through the rainbow once by setting rot=1 and reverse=True, etc. Furthermore, the
    algorithm was tuned, so that constant luminance values can be used (e.g. to create a truly
    isoluminant colorwheel). The `rot` parameter is also tuned to hold true for these cases.
    Of the here presented colorwheels, only this one manages to solely navigate through the L*=50
    plane, which can be seen here:
    https://upload.wikimedia.org/wikipedia/commons/2/21/Lab_color_space.png

    Parameters
    ----------
    start : scalar, optional
        Sets the starting position in the color space. 0=blue, 1=red,
        2=green. Defaults to 0.5.
    rot : scalar, optional
        The number of rotations through the rainbow. Can be positive
        or negative, indicating direction of rainbow. Negative values
        correspond to Blue->Red direction. Defaults to -1.5.
    gamma : scalar, optional
        The gamma correction for intensity. Defaults to 1.0.
    reverse : boolean, optional
        Set to True to reverse the color map. Will go from black to
        white. Good for density plots where shade~density. Defaults to False.
    nlev : scalar, optional
        Defines the number of discrete levels to render colors at.
        Defaults to 256.
    sat : scalar, optional
        The saturation intensity factor. Defaults to 1.2
        NOTE: this was formerly known as `hue` parameter
    minSat : scalar, optional
        Sets the minimum-level saturation. Defaults to 1.2.
    maxSat : scalar, optional
        Sets the maximum-level saturation. Defaults to 1.2.
    startHue : scalar, optional
        Sets the starting color, ranging from [0, 360], as in
        D3 version by @mbostock.
        NOTE: overrides values in start parameter.
    endHue : scalar, optional
        Sets the ending color, ranging from [0, 360], as in
        D3 version by @mbostock
        NOTE: overrides values in rot parameter.
    minLight : scalar, optional
        Sets the minimum lightness value. Defaults to 0.
    maxLight : scalar, optional
        Sets the maximum lightness value. Defaults to 1.

    Returns
    -------
    matplotlib.colors.LinearSegmentedColormap object

    Revisions
    ---------
    2014-04 (@jradavenport) Ported from IDL version
    2014-04 (@jradavenport) Added kwargs to enable similar to D3 version,
                            changed name of `hue` parameter to `sat`.
    2016-11 (@jan.caron) Added support for isoluminant cubehelices while making sure
                         `rot` works as intended. Decoded the plane-vectors a bit.

    """

    _log = logging.getLogger(__name__ + '.ColormapCubehelix')

    def __init__(self, start=0.5, rot=-1.5, gamma=1.0, reverse=False, nlev=256,
                 minSat=1.2, maxSat=1.2, minLight=0., maxLight=1., **kwargs):
        self._log.debug('Calling __init__')
        # Override start and rot if startHue and endHue are set:
        if kwargs is not None:
            if 'startHue' in kwargs:
                start = (kwargs.get('startHue') / 360. - 1.) * 3.
            if 'endHue' in kwargs:
                rot = kwargs.get('endHue') / 360. - start / 3. - 1.
            if 'sat' in kwargs:
                minSat = kwargs.get('sat')
                maxSat = kwargs.get('sat')
        self.nlev = nlev
        # Set up the parameters:
        self.fract = np.linspace(minLight, maxLight, nlev)
        angle = 2.0 * np.pi * (start / 3.0 + rot * np.linspace(0, 1, nlev))
        self.fract = self.fract**gamma
        satar = np.linspace(minSat, maxSat, nlev)
        amp = np.asarray(satar * self.fract * (1. - self.fract) / 2)
        # Set RGB color coefficients (Luma is calculated in RGB Rec.601, so choose those), the original version of
        # Dave Green used (0.30, 0.59, 0.11) and Rec.709 is c709 = (0.2126, 0.7152, 0.0722) but with eihter of those,
        # this function would not produce the correct YPbPr Luma.
        c601 = (0.299, 0.587, 0.114)
        cr, cg, cb = c601
        cw = -0.90649  # Chosen to comply with Dave Greens implementation.
        k = -1.6158 / cr / cw  # k has to balance out cw so nothing gets out of RGB gamut (> 1).
        # Calculate the vectors v and w spanning the plane of constant perceived intensity. v and w have to solve
        # v x w = k(cr, cg, cb) (normal vector of the described plane) and
        # v * w = 0 (scalar product, v and w have to be perpendicular).
        # 6 unknown and 4 equations --> Chose wb = 0 and wg = cw (constant).
        v = np.array((k * cr ** 2 * cb / (cw * (cr ** 2 + cg ** 2)),
                      k * cr * cg * cb / (cw * (cr ** 2 + cg ** 2)), -k * cr / cw))
        w = np.array((-cw * cg / cr, cw, 0))
        # Calculate components:
        self.red = self.fract + amp * (v[0] * np.cos(angle) + w[0] * np.sin(angle))
        self.grn = self.fract + amp * (v[1] * np.cos(angle) + w[1] * np.sin(angle))
        self.blu = self.fract + amp * (v[2] * np.cos(angle) + w[2] * np.sin(angle))
        # Original formulas with original v and w:
        # self.red = self.fract + amp * (-0.14861 * np.cos(angle) + 1.78277 * np.sin(angle))
        # self.grn = self.fract + amp * (-0.29227 * np.cos(angle) - 0.90649 * np.sin(angle))
        # self.blu = self.fract + amp * (1.97294 * np.cos(angle))
        # Find where RBG are outside the range [0,1], clip:
        self.red = np.clip(self.red, 0, 1)
        self.grn = np.clip(self.grn, 0, 1)
        self.blu = np.clip(self.blu, 0, 1)
        # Optional color reverse:
        if reverse is True:
            self.red = self.red[::-1]
            self.blu = self.blu[::-1]
            self.grn = self.grn[::-1]
        # Put in to tuple & dictionary structures needed:
        rr, bb, gg = [], [], []
        for k in range(0, int(nlev)):
            rr.append((float(k) / (nlev - 1), self.red[k], self.red[k]))
            bb.append((float(k) / (nlev - 1), self.blu[k], self.blu[k]))
            gg.append((float(k) / (nlev - 1), self.grn[k], self.grn[k]))
        cdict = {'red': rr, 'blue': bb, 'green': gg}
        super().__init__('cubehelix', cdict, N=256)
        self._log.debug('Created ' + str(self))

    def plot_helix(self, figsize=None, **kwargs):
        """Display the RGB and luminance plots for the chosen cubehelix.

        Parameters
        ----------
        figsize : tuple of floats (N=2)
            Size of the plot figure.

        Returns
        -------
        None

        """
        self._log.debug('Calling plot_helix')
        with use_style('empyre-plot'):
            fig = plt.figure(figsize=figsize, constrained_layout=True)
            gs = fig.add_gridspec(2, 1, height_ratios=[8, 1])
            # Main plot:
            axis = plt.subplot(gs[0])
            axis.plot(self.fract, 'k')
            axis.plot(self.red, 'r')
            axis.plot(self.grn, 'g')
            axis.plot(self.blu, 'b')
            axis.set_xlim(0, self.nlev)
            axis.set_ylim(0, 1)
            axis.set_title('Cubehelix')
            axis.set_xlabel('Color index')
            axis.set_ylabel('Brightness / RGB')
            axis.xaxis.set_major_locator(FixedLocator(locs=np.linspace(0, self.nlev, 5)))
            axis.yaxis.set_major_locator(FixedLocator(locs=[0, 0.5, 1]))
            # Colorbar horizontal:
            caxis = plt.subplot(gs[1])
            rgb = self(np.linspace(0, 1, 256))[None, ...]
            rgb = np.asarray(255.9999 * rgb, dtype=np.uint8)
            rgb = np.repeat(rgb, 30, axis=0)
            im = Image.fromarray(rgb)
            caxis.imshow(im, aspect='auto')
            caxis.tick_params(axis='both', which='both', labelleft=False, labelbottom=False,
                              left=False, right=False, top=False, bottom=False)


class ColormapPerception(colors.LinearSegmentedColormap, Colormap3D):
    """A perceptual colormap based on face-based luminance matching.

    Based on a publication by Kindlmann et. al.
    http://www.cs.utah.edu/~gk/papers/vis02/FaceLumin.pdf
    This colormap tries to achieve an isoluminant perception by using a list of colors acquired
    through face recognition studies. It is a lot better than the HLS colormap, but still not
    completely isoluminant (despite its name). Also it appears a bit dark.

    """

    _log = logging.getLogger(__name__ + '.ColormapPerception')

    CDICT = {'red': [(0/6, 0.847, 0.847),
                     (1/6, 0.527, 0.527),
                     (2/6, 0.000, 0.000),
                     (3/6, 0.000, 0.000),
                     (4/6, 0.316, 0.316),
                     (5/6, 0.718, 0.718),
                     (6/6, 0.847, 0.847)],

             'green': [(0/6, 0.057, 0.057),
                       (1/6, 0.527, 0.527),
                       (2/6, 0.592, 0.592),
                       (3/6, 0.559, 0.559),
                       (4/6, 0.316, 0.316),
                       (5/6, 0.000, 0.000),
                       (6/6, 0.057, 0.057)],

             'blue': [(0/6, 0.057, 0.057),
                      (1/6, 0.000, 0.000),
                      (2/6, 0.000, 0.000),
                      (3/6, 0.559, 0.559),
                      (4/6, 0.991, 0.991),
                      (5/6, 0.718, 0.718),
                      (6/6, 0.057, 0.057)]}

    def __init__(self):
        self._log.debug('Calling __init__')
        super().__init__('perception', self.CDICT, N=256)
        self._log.debug('Created ' + str(self))


class ColormapHLS(colors.ListedColormap, Colormap3D):
    """Colormap subclass for encoding directions with colors.

    This class is a subclass of the :class:`~matplotlib.pyplot.colors.ListedColormap`
    class. The class follows the HSL ('hue', 'saturation', 'lightness') 'Double Hexcone' Model
    with the saturation always set to 1 (moving on the surface of the color
    cylinder) with a lightness of 0.5 (full color). The three prime colors (`rgb`) are spaced
    equidistant with 120° space in between, according to a triadic arrangement.
    Even though the lightness is constant in the plane, the luminance (which is a weighted sum
    of the RGB components which encompasses human perception) is not, which can lead to
    artifacts like reliefs. Converting the map to a grayscale show spokes at the secondary colors.
    For more information see:
    https://vis4.net/blog/posts/avoid-equidistant-hsv-colors/
    http://www.workwithcolor.com/color-luminance-2233.htm
    http://blog.asmartbear.com/color-wheels.html

    """

    _log = logging.getLogger(__name__ + '.ColormapHLS')

    def __init__(self):
        self._log.debug('Calling __init__')
        h = np.linspace(0, 1, 256)
        l = 0.5 * np.ones_like(h)
        s = np.ones_like(h)
        r, g, b = np.vectorize(colorsys.hls_to_rgb)(h, l, s)
        colors = [(r[i], g[i], b[i]) for i in range(len(r))]
        super().__init__(colors, 'hls', N=256)
        self._log.debug('Created ' + str(self))


class ColormapClassic(colors.LinearSegmentedColormap, Colormap3D):
    """Colormap subclass for encoding directions with colors.

    This class is a subclass of the :class:`~matplotlib.pyplot.colors.LinearSegmentedColormap`
    class. The class follows the HSL ('hue', 'saturation', 'lightness') 'Double
    Hexcone' Model with the saturation always set to 1 (moving on the surface of the color
    cylinder) with a luminance of 0.5 (full color). The colors follow a tetradic arrangement with
    four colors (red, green, blue and yellow) arranged with 90° spacing in between.

    """

    _log = logging.getLogger(__name__ + '.ColormapClassic')

    CDICT = {'red': [(0/4, 1.0, 1.0),
                     (1/4, 0.0, 0.0),
                     (2/4, 0.0, 0.0),
                     (3/4, 1.0, 1.0),
                     (4/4, 1.0, 1.0)],

             'green': [(0/4, 0.0, 0.0),
                       (1/4, 0.0, 0.0),
                       (2/4, 1.0, 1.0),
                       (3/4, 1.0, 1.0),
                       (4/4, 0.0, 0.0)],

             'blue': [(0/4, 0.0, 0.0),
                      (1/4, 1.0, 1.0),
                      (2/4, 0.0, 0.0),
                      (3/4, 0.0, 0.0),
                      (4/4, 0.0, 0.0)]}

    def __init__(self):
        self._log.debug('Calling __init__')
        super().__init__('classic', self.CDICT, N=256)
        self._log.debug('Created ' + str(self))


class ColormapTransparent(colors.LinearSegmentedColormap):
    """Colormap subclass for including transparency.

    This class is a subclass of the :class:`~matplotlib.pyplot.colors.LinearSegmentedColormap`
    class with integrated support for transparency. The colormap is unicolor and varies only in
    transparency.

    Attributes
    ----------
    r: float, optional
        Intensity of red in the colormap. Has to be between 0. and 1.
    g: float, optional
        Intensity of green in the colormap. Has to be between 0. and 1.
    b: float, optional
        Intensity of blue in the colormap. Has to be between 0. and 1.
    alpha_range : list (N=2) of float, optional
        Start and end alpha value. Has to be between 0. and 1.

    """

    _log = logging.getLogger(__name__ + '.ColormapTransparent')

    def __init__(self, r=0., g=0., b=0., alpha_range=None):
        self._log.debug('Calling __init__')
        if alpha_range is None:
            alpha_range = [0., 1.]
        red = [(0., 0., r), (1., r, 1.)]
        green = [(0., 0., g), (1., g, 1.)]
        blue = [(0., 0., b), (1., b, 1.)]
        alpha = [(0., 0., alpha_range[0]), (1., alpha_range[1], 1.)]
        cdict = {'red': red, 'green': green, 'blue': blue, 'alpha': alpha}
        super().__init__('transparent', cdict, N=256)
        self._log.debug('Created ' + str(self))


def interpolate_color(fraction, start, end):
    """Interpolate linearly between two color tuples (e.g. RGB).

    Parameters
    ----------
    fraction: float or :class:`~numpy.ndarray`
        Interpolation fraction between 0 and 1, which determines the position of the
        interpolation between `start` and `end`.
    start: tuple (N=3) or :class:`~numpy.ndarray`
        Start of the interpolation as a tuple of three numbers or a numpy array, where the last
        dimension should have length 3 and contain the color tuples.
    end: tuple (N=3) or :class:`~numpy.ndarray`
        End of the interpolation as a tuple of three numbers or a numpy array, where the last
        dimension should have length 3 and contain the color tuples.

    Returns
    -------
    result: tuple (N=3) or :class:`~numpy.ndarray`
        Result of the interpolation as a tuple of three numbers or a numpy array, where the
        last dimension should has length 3 and contains the color tuples.

    """
    _log.debug('Calling interpolate_color')
    start, end = np.asarray(start), np.asarray(end)
    r1 = start[..., 0] + (end[..., 0] - start[..., 0]) * fraction
    r2 = start[..., 1] + (end[..., 1] - start[..., 1]) * fraction
    r3 = start[..., 2] + (end[..., 2] - start[..., 2]) * fraction
    return r1, r2, r3


cmaps = {'cubehelix_standard': ColormapCubehelix(),
         'cubehelix_reverse': ColormapCubehelix(reverse=True),
         'cubehelix_circular': ColormapCubehelix(start=1, rot=1, minLight=0.5, maxLight=0.5, sat=2),
         'perception_circular': ColormapPerception(),
         'hls_circular': ColormapHLS(),
         'classic_circular': ColormapClassic(),
         'transparent_black': ColormapTransparent(0, 0, 0, [0, 1.]),
         'transparent_white': ColormapTransparent(1, 1, 1, [0, 1.]),
         'transparent_confidence': ColormapTransparent(0.2, 0.3, 0.2, [0.75, 0.])}

CMAP_CIRCULAR_DEFAULT = cmaps['cubehelix_circular']



# -*- coding: utf-8 -*-
# Copyright 2019 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""This module provides a functions for 2D plots that often wrap functions from `maptlotlib.pyplot`."""


import logging

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from PIL import Image

from . import colors
from .tools import use_style
from ..fields.field import Field


__all__ = ['imshow', 'contour', 'colorvec', 'cosine_contours', 'quiver']

_log = logging.getLogger(__name__)


DIVERGING_CMAPS = ['PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu', 'RdYlBu', 'RdYlGn',
                   'Spectral', 'coolwarm', 'bwr', 'seismic',  # all divergent maps from matplotlib!
                   'balance', 'delta', 'curl', 'diff', 'tarn']  # all divergent maps from cmocean!
# TODO: add seaborn and more?


def imshow(field, axis=None, cmap=None, **kwargs):
    """Display an image on a 2D regular raster. Wrapper for `matplotlib.pyplot.imshow`.

    Parameters
    ----------
    field : `Field` or ndarray
        The image data as a `Field` or a numpy array (in the latter case, `vector=False` and `scale=1` are assumed).
    axis : `matplotlib.axes.Axes` object, optional
        The axis to which the image should be added, by default None, which will pick the last use axis via `gca`.
    cmap : str or `matplotlib.colors.Colormap`, optional
        The Colormap that should be used for the display, either as a string or object, by default None, which will pick
        `cmocean.cm.balance` if available. `imshow` will automatically detect if a divergent colormap is used and will
        make sure that zero is pinned to the symmetry point of the colormap (this is done by creating a new colormap
        with custom range under the hood).

    Returns
    -------
    axis : `matplotlib.axes.Axes`
        The plotting axis.

    Notes
    -----
    Additional kwargs are passed to `matplotlib.pyplot.imshow`.
    Uses the `empyre-image` stylesheet settings for plotting (and axis creation if none exists, yet).
    Fields are squeezed before plotting, so non-2D fields work as long as their superfluous dimensions have length 1.

    """
    _log.debug('Calling imshow')
    if not isinstance(field, Field):  # Try to convert input to Field if it is not already one:
        field = Field(data=np.asarray(field), scale=1, vector=False)
    assert not field.vector, 'Can only plot scalar fields!'
    # Get squeezed data and make sure it's 2D scalar:
    squeezed_field = field.squeeze()
    assert len(squeezed_field.dim) == 2, 'Cannot plot more than 2 dimensions!'
    # Determine colormap and related important properties and flags:
    if cmap is None:
        try:
            import cmocean
            cmap = cmocean.cm.balance
        except ImportError:
            _log.info('cmocean.balance not found, fallback to rRdBu!')
            cmap = plt.get_cmap('RdBu_r')  # '_r' for reverse!
    elif isinstance(cmap, str):  # make sure we have a Colormap object (and not a string):
        cmap = plt.get_cmap(cmap)
    if cmap.name.replace('_r', '') in DIVERGING_CMAPS:  # 'replace' also matches reverted cmaps!
        # Symmetric colormap only has zero at symmetry point if mappable has symmetric bounds (from - to + limit)!
        vmin = kwargs.get('vmin', squeezed_field.data.min()-1E-30)
        vmax = kwargs.get('vmax', squeezed_field.data.max()+1E-30)
        vmin, vmax = np.min([vmin, 0]), np.max([0, vmax])  # Ensure zero is present!
        kwargs.setdefault('vmin', vmin)
        kwargs.setdefault('vmax', vmax)
        limit = np.max(np.abs([vmin, vmax]))
        # Calculate the subset of colors for the range vmin to vmax (often not full range of 2*limit):
        start = 0.5 + vmin/(2*limit)  # 0 for symmetric bounds, >0: unused colors at lower end!
        end = 0.5 + vmax/(2*limit)  # 1 for symmetric bounds, <1: unused colors at upper end!
        cmap_colors = cmap(np.linspace(start, end, 256))  # New color indices with symmetry color at 0.5!
        # Use calculated colors to create custom (asymmetric) colormap with symmetry color (often white) at zero:
        cmap = LinearSegmentedColormap.from_list('custom', cmap_colors)
    # Set extent in data coordinates (left, right, bottom, top) to kwargs (if not set explicitely):
    dim_v, dim_u, s_v, s_u = *squeezed_field.dim, *squeezed_field.scale
    kwargs.setdefault('extent', (0, dim_u * s_u, 0, dim_v * s_v))
    # Plot with the empyre style context:
    with use_style('empyre-image'):  # Only works on axes created WITHIN context!
        if axis is None:  # If no axis is set, find the current or create a new one:
            axis = plt.gca()
        return axis.imshow(squeezed_field, cmap=cmap, **kwargs)


def contour(field, axis=None, **kwargs):
    """Plot contours. Wrapper for `matplotlib.pyplot.contour`.

    Parameters
    ----------
    field : `Field` or ndarray
        The contour data as a `Field` or a numpy array (in the latter case, `vector=False` and `scale=1` are assumed).
    axis : `matplotlib.axes.Axes` object, optional
        The axis to which the contour should be added, by default None, which will pick the last use axis via `gca`.

    Returns
    -------
    axis : `matplotlib.axes.Axes`
        The plotting axis.

    Notes
    -----
    Additional kwargs are passed to `matplotlib.pyplot.contour`.
    Uses the `empyre-image` stylesheet settings for plotting (and axis creation if none exists, yet).
    Fields are squeezed before plotting, so non-2D fields work as long as their superfluous dimensions have length 1.

    """
    _log.debug('Calling contour')
    if not isinstance(field, Field):  # Try to convert input to Field if it is not already one:
        field = Field(data=np.asarray(field), scale=1, vector=False)
    assert not field.vector, 'Can only plot scalar fields!'
    # Get squeezed data and make sure it's 2D scalar:
    squeezed_field = field.squeeze()
    assert len(squeezed_field.dim) == 2, 'Cannot plot more than 2 dimensions!'
    # Create coordinates (respecting the field scale, +0.5: pixel center!):
    vv, uu = (np.indices(squeezed_field.dim) + 0.5) * np.asarray(squeezed_field.scale)[:, None, None]
    # Set kwargs defaults without overriding possible user input:
    kwargs.setdefault('levels', [0.5])
    kwargs.setdefault('colors', 'k')
    kwargs.setdefault('linestyles', 'dotted')
    kwargs.setdefault('linewidths', 2)
    # Plot with the empyre style context:
    with use_style('empyre-image'):  # Only works on axes created WITHIN context!
        if axis is None:  # If no axis is set, find the current or create a new one:
            axis = plt.gca()
            axis.set_aspect('equal')
        return axis.contour(uu, vv, squeezed_field.data, **kwargs)


def colorvec(field, axis=None, **kwargs):
    """Plot an image of a 2D vector field with up to 3 components by color encoding the vector direction.

    In-plane directions are encoded via hue ("color wheel"), making sure that all in-plane directions are isoluminant
    (i.e. a greyscale image would result a homogeneously medium grey image). Out-of-plane directions are encoded via
    brightness with upwards pointing vectors being white and downward pointing vectors being black. The length of the
    vectors are encoded via saturation, with full saturation being fully chromatic (in-plane) or fully white/black
    (up/down). The center of the "color sphere" desaturated in a medium gray and encodes vectors with length zero.

    Parameters
    ----------
    field : `Field` or ndarray
        The image data as a `Field` or a numpy array (in the latter case, `vector=True` and `scale=1` are assumed).
    axis : `matplotlib.axes.Axes` object, optional
        The axis to which the image should be added, by default None, which will pick the last use axis via `gca`.

    Returns
    -------
    axis : `matplotlib.axes.Axes`
        The plotting axis.

    Notes
    -----
    Additional kwargs are passed to `matplotlib.pyplot.imshow`.
    Uses the `empyre-image` stylesheet settings for plotting (and axis creation if none exists, yet).
    Fields are squeezed before plotting, so non-2D fields work as long as their superfluous dimensions have length 1.
    Even though squeezing takes place, `colorvec` "remembers" the original orientation of the slice! This is important
    if you want to plot a slice that should not represent the xy-plane. The colors chosen will respect the original
    orientation of your slice, e.g. a vortex in the xz-plane will include black and white colors (up/down) if the
    `Field` object given as the `field` parameter has `dim=(128, 1, 128)`. If you want to plot a slice of a 3D vector
    with 3 components and make use of this functionality, make sure to not use an integer as an index, as that will
    drop the dimension BEFORE it is passed to `colorvec`, which will have no way of knowing which dimension was dropped.
    Instead, make sure to use a slice of length one (example with `dim=(128, 128, 128)`):
    >>> colorvec(field[:, 15, :])  # Wrong! Shape: (128, 128), interpreted as xy-plane!
    >>> colorvec(field[:, 15:16, :])  # Right! Shape: (128, 1, 128), passed as 3D to `colorvec`, squeezed internally!

    """
    _log.debug('Calling colorvec')
    if not isinstance(field, Field):  # Try to convert input to Field if it is not already one:
        field = Field(data=np.asarray(field), scale=1, vector=True)
    assert field.vector, 'Can only plot vector fields!'
    assert len(field.dim) <= 3, 'Unusable for vector fields with dimension higher than 3!'
    assert len(field.dim) == field.ncomp, ('Assignment of vector components to dimensions is ambiguous!'
                                           f'`ncomp` ({field.ncomp}) must match `len(dim)` ({len(field.dim)})')
    # Get squeezed data and make sure it's 2D scalar:
    squeezed_field = field.squeeze()
    assert len(squeezed_field.dim) == 2, 'Cannot plot more than 2 dimensions!'
    # Extract vector components (fill 3rd component with zeros if field.comp is only 2):
    comp = squeezed_field.comp
    x_comp = comp[0]
    y_comp = comp[1]
    z_comp = comp[2] if (squeezed_field.ncomp == 3) else np.zeros(squeezed_field.dim)
    # Calculate image with color encoded directions:
    rgb = colors.CMAP_CIRCULAR_DEFAULT.rgb_from_vector(np.stack((x_comp, y_comp, z_comp), axis=0))
    # Set extent in data coordinates (left, right, bottom, top) to kwargs (if not set explicitely):
    dim_v, dim_u, s_v, s_u = *squeezed_field.dim, *squeezed_field.scale
    kwargs.setdefault('extent', (0, dim_u * s_u, 0, dim_v * s_v))
    # Plot with the empyre style context:
    with use_style('empyre-image'):  # Only works on axes created WITHIN context!
        if axis is None:  # If no axis is set, find the current or create a new one:
            axis = plt.gca()
        return axis.imshow(Image.fromarray(rgb), **kwargs)


def cosine_contours(field, axis=None, gain='auto', cmap=None, **kwargs):
    """Plots the cosine of the (amplified) field. Wrapper for `matplotlib.pyplot.imshow`.

    Parameters
    ----------
    field : `Field` or ndarray
        The contour data as a `Field` or a numpy array (in the latter case, `vector=False` and `scale=1` are assumed).
    axis : `matplotlib.axes.Axes` object, optional
        The axis to which the contour should be added, by default None, which will pick the last use axis via `gca`.
    gain : float or 'auto', optional
        Gain factor with which the `Field` is amplified before taking the cosine, by default 'auto', which calculates a
        gain factor that would produce roughly 4 cosine contours.
    cmap : str or `matplotlib.colors.Colormap`, optional
        The Colormap that should be used for the display, either as a string or object, by default None, which will pick
        `colors.cmaps['transparent_black']` that will alternate between regions with alpha=0, showing layers below and
        black contours.

    Returns
    -------
    axis : `matplotlib.axes.Axes`
        The plotting axis.

    Notes
    -----
    Additional kwargs are passed to `matplotlib.pyplot.imshow`.
    Uses the `empyre-image` stylesheet settings for plotting (and axis creation if none exists, yet).
    Fields are squeezed before plotting, so non-2D fields work as long as their superfluous dimensions have length 1.

    """
    _log.debug('Calling cosine_contours')
    if not isinstance(field, Field):  # Try to convert input to Field if it is not already one:
        field = Field(data=np.asarray(field), scale=1, vector=False)
    assert not field.vector, 'Can only plot scalar fields!'
    # Get squeezed data and make sure it's 2D scalar:
    squeezed_field = field.squeeze()
    assert len(squeezed_field.dim) == 2, 'Cannot plot more than 2 dimensions (Squeezing did not help)!'
    # Determine colormap and related important properties and flags:
    if cmap is None:
        cmap = colors.cmaps['transparent_black']
    # Calculate gain if 'auto' is selected:
    if gain == 'auto':
        gain = 4 * 2*np.pi / (squeezed_field.amp.data.max() + 1E-30)  # 4: roughly 4 contours!
        gain = round(gain, -int(np.floor(np.log10(abs(gain)))))  # Round to last significant digit!
        _log.info(f'Automatically calculated a gain of: {gain}')
    # Calculate the contours:
    contours = np.cos(gain * squeezed_field)  # Range: [-1, 1]
    contours += 1  # Shift to positive values
    contours /= 2  # Rescale to [0, 1]
    # Set extent in data coordinates (left, right, bottom, top) to kwargs (if not set explicitely):
    dim_v, dim_u, s_v, s_u = *squeezed_field.dim, *squeezed_field.scale
    kwargs.setdefault('extent', (0, dim_u * s_u, 0, dim_v * s_v))
    # Plot with the empyre style context:
    with use_style('empyre-image'):  # Only works on axes created WITHIN context!
        if axis is None:  # If no axis is set, find the current or create a new one:
            axis = plt.gca()
        return axis.imshow(contours, cmap=cmap, **kwargs)


def quiver(field, axis=None, color_angles=False, cmap=None, n_bin='auto', bin_with_mask=True, **kwargs):
    """Plot a 2D field of arrows. Wrapper for `matplotlib.pyplot.imshow`.

    Parameters
    ----------
    field : `Field` or ndarray
        The vector data as a `Field` or a numpy array (in the latter case, `vector=True` and `scale=1` are assumed).
    axis : `matplotlib.axes.Axes` object, optional
        The axis to which the image should be added, by default None, which will pick the last use axis via `gca`.
    color_angles : bool, optional
        Switch that turns on color encoding of the arrows, by default False. Encoding works the same as for the
        `colorvec` function (see for details). If False, arrows are uniformly colored white with black border. In both
        cases, the amplitude is encoded via the transparency of the arrow.
    cmap : str or `matplotlib.colors.Colormap`, optional
        The Colormap that should be used for the arrows, either as a string or object, by default None. Will only be
        used if `color_angles=True`.
    n_bin : float or 'auto', optional
        Number of entries along each axis over which the average is taken, by default 'auto', which automatically
        determines a bin size resulting in roughly 16 arrows along the largest dimension. Usually sensible to leave
        this on to not clutter the image with too many arrows (also due to performance). Can be turned off by setting
        `n_bin=1`. Uses the `..fields.field.Field.bin` method.
    bin_with_mask : bool, optional
        If True (default) and if `n_bin>1`, entries of the constructed binned `Field` that averaged over regions that
        were outside the `..fields.field.Field.mask` will not be assigned an arrow and stay empty instead. This prevents
        errouneous "fade-out" effects of the arrows that would occur even for homogeneous objects.

    Returns
    -------
    quiv : Quiver instance
        The quiver instance that was created.

    Notes
    -----
    Additional kwargs are passed to `matplotlib.pyplot.quiver`.
    Uses the `empyre-image` stylesheet settings for plotting (and axis creation if none exists, yet).
    Fields are squeezed before plotting, so non-2D fields work as long as their superfluous dimensions have length 1.
    Even though squeezing takes place, `quiver` "remembers" the original orientation of the slice and which dimensions
    were squeezed! See `colorvec` for  more information and an example (the same principles apply here, too).
    The transparency of the arrows denotes the 3D(!) amplitude, if you see dots in the plot, that means the amplitude
    is not zero, but simply out of the current plane!

    """
    _log.debug('Calling quiver')
    if not isinstance(field, Field):  # Try to convert input to Field if it is not already one:
        field = Field(data=np.asarray(field), scale=1, vector=True)
    assert field.vector, 'Can only plot vector fields!'
    assert len(field.dim) <= 3, 'Unusable for vector fields with dimension higher than 3!'
    assert len(field.dim) == field.ncomp, ('Assignment of vector components to dimensions is ambiguous!'
                                           f'`ncomp` ({field.ncomp}) must match `len(dim)` ({len(field.dim)})')
    # Get squeezed data and make sure it's 2D scalar:
    squeezed_field = field.squeeze()
    assert len(squeezed_field.dim) == 2, 'Cannot plot more than 2 dimensions (Squeezing did not help)!'
    # Determine binning size if necessary:
    if n_bin == 'auto':
        n_bin = int(np.max((1, np.max(squeezed_field.dim) / 16)))
    # Save old limits in case binning has to use padding:
    u_lim = squeezed_field.dim[1] * squeezed_field.scale[1]
    v_lim = squeezed_field.dim[0] * squeezed_field.scale[0]
    # Bin if necessary:
    if n_bin > 1:
        field_mask = squeezed_field.mask  # Get mask BEFORE binning!
        squeezed_field = squeezed_field.bin(n_bin)
        if bin_with_mask:  # Excludes regions where in and outside are binned together!
            mask = (field_mask.bin(n_bin) == 1)
            squeezed_field *= mask
    # Extract normalized vector components (fill 3rd component with zeros if field.comp is only 2):
    normalised_comp = (squeezed_field / squeezed_field.amp.data.max()).comp
    amplitude = squeezed_field.amp.data / squeezed_field.amp.data.max()
    x_comp = normalised_comp[0].data
    y_comp = normalised_comp[1].data
    z_comp = normalised_comp[2].data if (field.ncomp == 3) else np.zeros(squeezed_field.dim)
    # Create coordinates (respecting the field scale, +0.5: pixel center!):
    vv, uu = (np.indices(squeezed_field.dim) + 0.5) * np.asarray(squeezed_field.scale)[:, None, None]
    # Calculate the arrow colors:
    if color_angles:  # Color angles according to calculated RGB values (only with circular colormaps):
        _log.debug('Encoding angles')
        if cmap is None:
            cmap = colors.CMAP_CIRCULAR_DEFAULT
        rgb = cmap.rgb_from_vector(np.asarray((x_comp, y_comp, z_comp))) / 255
        rgba = np.concatenate((rgb, amplitude[..., None]), axis=-1)
        kwargs.setdefault('color', rgba.reshape(-1, 4))
    else:  # Color amplitude with numeric values, according to cmap, overrides 'color':
        _log.debug('Encoding amplitudes')
        if cmap is None:
            cmap = colors.cmaps['transparent_white']
        C = amplitude  # Numeric values, used with cmap!
    # Check which (if any) indices were squeezed to find out which components are passed to quiver:  # TODO: TEST!!!
    squeezed_indices = np.flatnonzero(np.asarray(field.dim) == 1)
    if not squeezed_indices:  # Separate check, because in this case squeezed_indices == []:
        u_comp = x_comp
        v_comp = y_comp
    elif squeezed_indices[0] == 0:  # Slice of the xy-plane with z squeezed:
        u_comp = x_comp
        v_comp = y_comp
    elif squeezed_indices[0] == 1:  # Slice of the xz-plane with y squeezed:
        u_comp = x_comp
        v_comp = z_comp
    elif squeezed_indices[0] == 2:  # Slice of the zy-plane with x squeezed:
        u_comp = y_comp
        v_comp = z_comp
    # Set specific defaults for quiver kwargs:
    kwargs.setdefault('edgecolor', colors.cmaps['transparent_black'](amplitude).reshape(-1, 4))
    kwargs.setdefault('scale', 1/np.max(squeezed_field.scale))
    kwargs.setdefault('width', np.max(squeezed_field.scale))
    kwargs.setdefault('clim', (0, 1))
    kwargs.setdefault('pivot', 'middle')
    kwargs.setdefault('units', 'xy')
    kwargs.setdefault('scale_units', 'xy')
    kwargs.setdefault('minlength', 0.05)
    kwargs.setdefault('headlength', 2)
    kwargs.setdefault('headaxislength', 2)
    kwargs.setdefault('headwidth', 2)
    kwargs.setdefault('minshaft', 2)
    kwargs.setdefault('linewidths', 1)
    # Plot with the empyre style context:
    with use_style('empyre-image'):  # Only works on axes created WITHIN context!
        if axis is None:  # If no axis is set, find the current or create a new one:
            axis = plt.gca()
        axis.set_xlim(0, u_lim)
        axis.set_ylim(0, v_lim)
        axis.set_aspect('equal')
        if color_angles:
            return axis.quiver(uu, vv, np.asarray(u_comp), np.asarray(v_comp), cmap=cmap, **kwargs)
        else:
            return axis.quiver(uu, vv, np.asarray(u_comp), np.asarray(v_comp), C, cmap=cmap, **kwargs)



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""This module provides functions that decorate exisiting `matplotlib` plots."""


import logging
from collections.abc import Iterable

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import patheffects
from matplotlib.patches import Circle
from matplotlib.offsetbox import TextArea, AnchoredOffsetbox
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from . import colors
from .tools import use_style
from ..fields.field import Field


__all__ = ['scalebar', 'colorwheel', 'annotate', 'quiverkey', 'coords', 'colorbar']

_log = logging.getLogger(__name__)


def scalebar(axis=None, unit='nm', loc='lower left', **kwargs):
    """Add a scalebar to the axis.

    Parameters
    ----------
    axis : :class:`~matplotlib.axes.AxesSubplot`, optional
        Axis to which the scalebar is added, by default None, which will pick the last used axis via `gca`.
    unit: str, optional
        String that determines the unit of the scalebar, defaults to 'nm'.
    loc : str or pair of floats, optional
        The location of the scalebar, defaults to 'lower left'. See `matplotlib.legend` for possible settings.

    Returns
    -------
    aoffbox : :class:`~matplotlib.offsetbox.AnchoredOffsetbox`
        The box containing the scalebar.

    Notes
    -----
    Additional kwargs are passed to `mpl_toolkits.axes_grid1.anchored_artists.AnchoredSizeBar`.

    """
    _log.debug('Calling scalebar')
    if axis is None:  # If no axis is set, find the current or create a new one:
        axis = plt.gca()
    # Transform axis borders (1, 1) to data borders to get number of pixels in y and x:
    transform = axis.transData
    bb0 = axis.transLimits.inverted().transform((0, 0))
    bb1 = axis.transLimits.inverted().transform((1, 1))
    data_extent = (int(abs(bb1[1] - bb0[1])), int(abs(bb1[0] - bb0[0])))
    # Calculate bar length:
    bar_length = data_extent[1] / 4  # 25% of the data width!
    thresholds = [1, 5, 10, 50, 100, 500, 1000]
    for t in thresholds:  # For larger grids (real images), multiples of threshold look better!
        if bar_length > t:  # Round down to the next lowest multiple of t:
            bar_length = (bar_length // t) * t
    # Set parameters for scale bar:
    label = f'{bar_length:g} {unit}'
    # Set defaults:
    kwargs.setdefault('borderpad', 0.2)
    kwargs.setdefault('pad', 0.2)
    kwargs.setdefault('sep', 5)
    kwargs.setdefault('color', 'w')
    kwargs.setdefault('size_vertical', data_extent[0]*0.01)
    kwargs.setdefault('frameon', False)
    kwargs.setdefault('label_top', True)
    kwargs.setdefault('fill_bar', True)
    # Create scale bar:
    scalebar = AnchoredSizeBar(transform, bar_length, label, loc, **kwargs)
    scalebar.txt_label._text._color = 'w'  # Overwrite AnchoredSizeBar color!
    # Set stroke patheffect:
    effect_txt = [patheffects.withStroke(linewidth=2, foreground='k')]
    scalebar.txt_label._text.set_path_effects(effect_txt)
    effect_bar = [patheffects.withStroke(linewidth=3, foreground='k')]
    scalebar.size_bar._children[0].set_path_effects(effect_bar)
    # Add scale bar to axis and return:
    axis.add_artist(scalebar)
    return scalebar


def colorwheel(axis=None, cmap=None, ax_size='20%', loc='upper right', **kwargs):
    """Add a colorwheel to the axis on the upper right corner.

    Parameters
    ----------
    axis : :class:`~matplotlib.axes.Axes`, optional
        Axis to which the colorwheel is added, by default None, which will pick the last used axis via `gca`.
    cmap : str or `matplotlib.colors.Colormap`, optional
        The Colormap that should be used for the colorwheel, defaults to `None`, which chooses a map specified by
        `.colors.CMAP_CIRCULAR_DEFAULT`. Needs to be a :class:`~.colors.Colormap3D`.
    ax_size : str or float, optional
        String or float determining the size of the inset axis used, defaults to `20%`.
    loc : str or pair of floats, optional
        The location of the colorwheel, defaults to 'upper right'. See `matplotlib.legend` for possible settings.

    Returns
    -------
    axis : :class:`~matplotlib.image.AxesImage`
        The colorwheel image that was created.

    Notes
    -----
    Additional kwargs are passed to :class:`~.colors.Colormap3D.plot_colorwheel` of the :class:`~.colors.Colormap3D`.

    """
    _log.debug('Calling colorwheel')
    if axis is None:  # If no axis is set, find the current or create a new one:
        axis = plt.gca()
    ins_axes = inset_axes(axis, width=ax_size, height=ax_size, loc=loc)
    ins_axes.axis('off')
    if cmap is None:
        cmap = colors.CMAP_CIRCULAR_DEFAULT
    plt.sca(axis)  # Set focus back to parent axis!
    return cmap.plot_colorwheel(axis=ins_axes, **kwargs)


def annotate(label, axis=None, loc='upper left'):
    """Add an annotation to the axis on the upper left corner.

    Parameters
    ----------
    label : string
        The text of the annotation.
    axis : :class:`~matplotlib.axes.AxesSubplot`, optional
        Axis to which the annotation is added, by default None, which will pick the last used axis via `gca`.
    loc : str or pair of floats, optional
        The location of the annotation, defaults to 'upper left'. See `matplotlib.legend` for possible settings.

    Returns
    -------
    aoffbox : :class:`~matplotlib.offsetbox.AnchoredOffsetbox`
        The box containing the annotation.

    """
    _log.debug('Calling annotate')
    if axis is None:  # If no axis is set, find the current or create a new one:
        axis = plt.gca()
    # Create text:
    txt = TextArea(label, textprops={'color': 'w'})
    txt.set_clip_box(axis.bbox)
    txt._text.set_path_effects([patheffects.withStroke(linewidth=2, foreground='k')])
    # Pack into and add AnchoredOffsetBox:
    aoffbox = AnchoredOffsetbox(loc=loc, pad=0.5, borderpad=0.1, child=txt, frameon=False)
    axis.add_artist(aoffbox)
    return aoffbox


def quiverkey(quiv, field, axis=None, unit='', loc='lower right', **kwargs):
    """Add a quiver key to an axis.

    Parameters
    ----------
    quiv : Quiver instance
        The quiver instance returned by a call to quiver.
    field : `Field` or ndarray
        The vector data as a `Field` or a numpy array (in the latter case, `vector=True` and `scale=1` are assumed).
    axis : :class:`~matplotlib.axes.AxesSubplot`, optional
        Axis to which the quiverkey is added, by default None, which will pick the last used axis via `gca`.
    unit: str, optional
        String that determines the unit of the quiverkey, defaults to ''.
    loc : str or pair of floats, optional
        The location of the quiverkey, defaults to 'lower right'. See `matplotlib.legend` for possible settings.

    Returns
    -------
    qk: Quiverkey
        The generated quiverkey.

    Notes
    -----
    Additional kwargs are passed to `matplotlib.pyplot.quiverkey`.

    """
    _log.debug('Calling quiverkey')
    if axis is None:  # If no axis is set, find the current or create a new one:
        axis = plt.gca()
    if not isinstance(field, Field):  # Try to convert input to Field if it is not already one:
        field = Field(data=np.asarray(field), scale=1, vector=True)
    length = field.amp.data.max()
    shift = 1 / field.squeeze().dim[1]  # equivalent to one pixel distance in axis coords!
    label = f'{length:.3g} {unit}'
    if loc in ('upper right', 1):
        X, Y, labelpos = 0.95-shift, 0.95-shift/4, 'W'
    elif loc in ('upper left', 2):
        X, Y, labelpos = 0.05+shift, 0.95-shift/4, 'E'
    elif loc in ('lower left', 3):
        X, Y, labelpos = 0.05+shift, 0.05+shift/4, 'E'
    elif loc in ('lower right', 4):
        X, Y, labelpos = 0.95-shift, 0.05+shift/4, 'W'
    else:
        raise ValueError('Quiverkey can only be placed in one of the corners (number 1 - 4 or associated strings)!')
    # Set defaults:
    kwargs.setdefault('coordinates', 'axes')
    kwargs.setdefault('facecolor', 'w')
    kwargs.setdefault('edgecolor', 'k')
    kwargs.setdefault('labelcolor', 'w')
    kwargs.setdefault('linewidth', 1)
    kwargs.setdefault('clip_box', axis.bbox)
    kwargs.setdefault('clip_on', True)
    # Plot:
    qk = axis.quiverkey(quiv, X, Y, U=1, label=label, labelpos=labelpos, **kwargs)
    qk.text.set_path_effects([patheffects.withStroke(linewidth=2, foreground='k')])
    return qk


def coords(axis=None, coords=('x', 'y'), loc='lower left', **kwargs):
    """Add coordinate arrows to an axis.

    Parameters
    ----------
    axis : :class:`~matplotlib.axes.AxesSubplot`, optional
        Axis to which the coordinates are added, by default None, which will pick the last used axis via `gca`.
    coords : tuple or int, optional
        Tuple of strings determining the labels, by default ('x', 'y'). Can also be `2` or `3` which expands to
        ('x', 'y') or ('x', 'y', 'z'). The length of `coords` determines the number of arrows (2 or 3).
    loc : str, optional
        [description], by default 'lower left'

    Returns
    -------
    ins_axes : :class:`~matplotlib.axes.Axes`
        The created inset axes containing the coordinates.

    """
    _log.debug('Calling coords')
    if axis is None:  # If no axis is set, find the current or create a new one:
        axis = plt.gca()
    ins_ax = inset_axes(axis, width='5%', height='5%', loc=loc, borderpad=2.2)
    if coords == 3:
        coords = ('x', 'y', 'z')
    elif coords == 2:
        coords = ('x', 'y')
    effects = [patheffects.withStroke(linewidth=2, foreground='k')]
    kwargs.setdefault('fc', 'w')
    kwargs.setdefault('ec', 'k')
    kwargs.setdefault('head_width', 0.6)
    kwargs.setdefault('head_length', 0.7)
    kwargs.setdefault('linewidth', 1)
    kwargs.setdefault('width', 0.2)
    if len(coords) == 3:
        ins_ax.arrow(x=0.5, y=0.5, dx=-1.05, dy=-0.75, clip_on=False, **kwargs)
        ins_ax.arrow(x=0.5, y=0.5, dx=0.96, dy=-0.75, clip_on=False, **kwargs)
        ins_ax.arrow(x=0.5, y=0.5, dx=0, dy=1.35, clip_on=False, **kwargs)
        ins_ax.annotate(coords[0], xy=(0, 0), xytext=(-1.0, 0.3), path_effects=effects, color='w')
        ins_ax.annotate(coords[1], xy=(0, 0), xytext=(1.7, 0.3), path_effects=effects, color='w')
        ins_ax.annotate(coords[2], xy=(0, 0), xytext=(0.8, 1.5), path_effects=effects, color='w')
        ins_ax.add_artist(Circle((0.5, 0.5), 0.12, fc='w', ec='k', linewidth=1, clip_on=False))
    elif len(coords) == 2:
        ins_ax.arrow(x=-0.5, y=-0.5, dx=1.5, dy=0, clip_on=False, **kwargs)
        ins_ax.arrow(x=-0.5, y=-0.5, dx=0, dy=1.5, clip_on=False, **kwargs)
        ins_ax.annotate(coords[0], xy=(0, 0), xytext=(1.3, -0.05), path_effects=effects, color='w')
        ins_ax.annotate(coords[1], xy=(0, 0), xytext=(-0.1, 1.1), path_effects=effects, color='w')
        ins_ax.add_artist(Circle((-0.5, -0.5), 0.12, fc='w', ec='k', linewidth=1, clip_on=False))
    ins_ax.axis('off')
    plt.sca(axis)
    return coords


def colorbar(im, fig=None, cbar_axis=None, axes=None, position='right', pad=0.02, thickness=0.03, label=None,
             constrain_ticklabels=True, ticks=None, ticklabels=None):
    """Creates a colorbar, aligned with figure axes.

    Parameters
    ----------
    im : matplotlib object, mappable
        Mappable matplotlib object.
    fig : matplotlib.figure object, optional
        The figure object that contains the matplotlib axes and artists, by default None, which will pick the last used
        figure via `gcf`.
    axes : matplotlib.axes or list of matplotlib.axes
        The axes object(s), where the colorbar is drawn, by default None, which will pick the last used axis via `gca`.
        Only provide those axes, which the colorbar should span over.
    position : str, optional
        The position defines the location of the colorbar. One of 'top', 'bottom', 'left' or 'right' (default).
    pad : float, optional
        Defines the spacing between the axes and colorbar axis. Is given in figure fraction.
    thickness : float, optional
        Thickness of the colorbar given in figure fraction.
    label : string, optional
        Colorbar label, defaults to None.
    constrain_ticklabels : bool, optional
        Allows to slightly shift the outermost ticklabels, such that they do not exceed the cbar axis, defaults to True.
    ticks : list, np.ndarray, optional
        List of cbar ticks, defaults to None.
    ticklabels : list, np.ndarray, optional
        List of cbar ticklabels, defaults to None.

    Returns
    -------
    cbar : :class:`~matplotlib.Colorbar`
        The created colorbar.

    Notes
    -----
    Based on a modified snippet by Florian Winkler. Note that this function TURNS OFF constrained layout, therefore it
    should be the final command before finishing or saving a figure. The colorbar will be outside the original bounds
    of your constructed figure. If you set the size, e.g. with `~empyre.vis.tools.new`, make sure to account for the
    additional space by setting the `width_scale` to something smaller than 1 (e.g. 0.9).

    """
    _log.debug('Calling colorbar')
    assert position in ('left', 'right', 'top', 'bottom'), "position has to be 'left', 'right', 'top' or 'bottom'!"
    if fig is None:  # If no figure is set, find the current or create a new one:
        fig = plt.gcf()
    fig.canvas.draw()  # Trigger a draw so that a potential constrained_layout is executed once!
    fig.set_constrained_layout(False)  # we don't want the layout to change after this point!
    if axes is None:  # If no axis is set, find the current or create a new one:
        axes = plt.gca()
    if not isinstance(axes, Iterable):
        axes = (axes,)  # Make sure it is an iterable (e.g. tuple)!
    # Save previously active axis for later:
    previous_axis = plt.gca()
    if cbar_axis is None:  # Construct a new cbar_axis:
        x_coords, y_coords = [], []
        # Find bounds of all individual axes:
        for ax in np.ravel(axes):  # ravel needed for arrays of axes:
            points = ax.get_position().get_points()
            x_coords.extend(points[:, 0])
            y_coords.extend(points[:, 1])
        # Find outer bounds of plotting area:
        left, right = min(x_coords), max(x_coords)
        bottom, top = min(y_coords), max(y_coords)
        # Determine where the colorbar will be placed:
        if position == 'right':
            bounds = [right+pad, bottom, thickness, top-bottom]
        elif position == 'left':
            bounds = [left-pad-thickness, bottom, thickness, top-bottom]
        if position == 'top':
            bounds = [left, top+pad, right-left, thickness]
        elif position == 'bottom':
            bounds = [left, bottom-pad-thickness, right-left, thickness]
        cbar_axis = fig.add_axes(bounds)
    # Create the colorbar:
    with use_style('empyre-image'):
        if position in ('left', 'right'):
            cb = plt.colorbar(im, cax=cbar_axis, orientation='vertical')
            cb.ax.yaxis.set_ticks_position(position)
            cb.ax.yaxis.set_label_position(position)
        elif position in ('top', 'bottom'):
            cb = plt.colorbar(im, cax=cbar_axis, orientation='horizontal')
            cb.ax.xaxis.set_ticks_position(position)
            cb.ax.xaxis.set_label_position(position)
        # Colorbar label
        if label is not None:
            cb.set_label(f'{label}')
        # Set ticks and ticklabels (if specified):
        if ticks:
            cb.set_ticks(ticks)
        if ticklabels:
            cb.set_ticklabels(ticklabels)
        # Constrain tick labels (if wanted):
        if constrain_ticklabels:
            if position == 'top' or position == 'bottom':
                t = cb.ax.get_xticklabels()
                t[0].set_horizontalalignment('left')
                t[-1].set_horizontalalignment('right')
            elif position == 'left' or position == 'right':
                t = cb.ax.get_yticklabels()
                t[0].set_verticalalignment('bottom')
                t[-1].set_verticalalignment('top')
    # Set focus back from colorbar to previous axis and return colorbar:
    plt.sca(previous_axis)
    return cb



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#


import logging

import numpy as np

from .field import Field


__all__ = ['create_shape_slab', 'create_shape_disc', 'create_shape_ellipse', 'create_shape_ellipsoid',
           'create_shape_sphere', 'create_shape_filament', 'create_shape_voxel']

_log = logging.getLogger(__name__)


# TODO: TEST!!!


def create_shape_slab(dim, center=None, width=None, scale=1):
    """Creates a Field object with the shape of a slab as a scalar field in arbitrary dimensions.

    Attributes
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple, optional
        The center of the slab in pixel coordinates.
    width : tuple, optional
        The width of the slab in pixel coordinates.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    if center is None:
        center = tuple([d/2 for d in dim])
    if width is None:
        width = tuple([d/2 for d in dim])
    assert len(dim) == len(center) == len(width), 'Parameters dim, center and width must have the same dimensions!'
    data = np.zeros(dim)
    bounds = ()
    for i in range(len(dim)):
        start = int(np.floor(center[i] - width[i]/2))
        stop = int(np.ceil(center[i] + width[i]/2))
        bounds += (slice(start, stop),)
    data[bounds] = 1
    return Field(data=data, scale=scale, vector=False)


def create_shape_disc(dim, center=None, radius=None, height=None, axis=0, scale=1):
    """Creates a Field object with the shape of a cylindrical disc in 2D or 3D.

    Attributes
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple, optional
        The center of the disc in pixel coordinates.
    radius : float, optional
        The radius of the disc in pixel coordinates.
    height : float, optional
        The height of the disc in pixel coordinates. Unused if only 2D.
    axis : int, optional
        The orientation of the discs orthogonal axis. Only used in 3D case with z-axis as default.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    assert len(dim) in (2, 3), 'Disc can only be build in 2 or 3 dimensions!'
    # Find indices of the disc plane axes:
    idx_uv = [0, 1, 2]
    if len(dim) == 3:  # 3D:
        idx_uv.remove(axis)
    else:  # 2D:
        idx_uv.remove(2)
    # Find default values:
    if center is None:
        center = tuple([d/2 for d in dim])
    if radius is None:
        radius = np.max((dim[idx_uv[0]], dim[idx_uv[1]])) / 4
    if height is None and len(dim) == 3:  # only used for 3D!
        height = dim[axis] / 2
        assert height > 0, 'Height has to be a positive scalar value!'
    assert len(dim) == len(center), 'Parameters dim and center must have the same dimensions!'
    assert radius > 0, 'Radius has to be a positive scalar value!'
    coords = np.indices(dim) + 0.5  # 0.5 to get to pixel/voxel center!
    coords = coords - np.asarray(center)[(slice(None),) + (None,)*len(dim)]  # [:, None, None, None]
    data = np.where(np.hypot(coords[idx_uv[0]], coords[idx_uv[1]]) <= radius, 1, 0)
    if len(dim) == 3:  # 3D: Implement bounds above and below the disc:
        height_shape = np.zeros_like(data)
        bounds = [slice(None)] * 3  # i.e.: [:, :, :]
        start = int(np.floor(center[axis] - height/2))
        stop = int(np.ceil(center[axis] + height/2))
        bounds[axis] = slice(start, stop)  # replace with actual bounds along disc symmetry axis
        height_shape[tuple(bounds)] = 1
        data *= height_shape
    return Field(data=data, scale=scale, vector=False)


def create_shape_ellipse(dim, center=None, width=None, height=None, axis=0, scale=1):
    """Creates a Field object with the shape of an ellipse in 2D or 3D.

    Attributes
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple, optional
        The center of the ellipse in pixel coordinates.
    width : tuple, optional
        The two half axes of the ellipse in pixel coordinates.
    height : float, optional
        The height of the ellipse in pixel coordinates. Unused if only 2D.
    axis : int, optional
        The orientation of the ellipses orthogonal axis. Only used in 3D case with z-axis as default.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    assert len(dim) in (2, 3), 'Ellipse can only be build in 2 or 3 dimensions!'
    # Find indices of the disc plane axes:
    idx_uv = [0, 1, 2]
    if len(dim) == 3:  # 3D:
        idx_uv.remove(axis)
    else:  # 2D:
        idx_uv.remove(2)
    # Find default values:
    if center is None:
        center = tuple([d/2 for d in dim])
    if width is None:
        dim_uv = (dim[idx_uv[0]], dim[idx_uv[1]])
        width = (np.max(dim_uv)*1/3, np.min(dim_uv)*2/3)
    if height is None and len(dim) == 3:  # only used for 3D!
        height = dim[axis] / 2
    assert len(dim) == len(center), 'Parameters dim and center must have the same dimensions!'
    assert len(width) == 2, 'Width has to contain the two half axes!'
    coords = np.indices(dim) + 0.5  # 0.5 to get to pixel/voxel center!
    coords = coords - np.asarray(center)[(slice(None),) + (None,)*len(dim)]  # [:, None, None, None]
    distance = np.hypot(coords[idx_uv[1]] / (width[1]/2), coords[idx_uv[0]] / (width[0]/2))
    data = np.where(distance <= 1, 1, 0)
    if len(dim) == 3:  # 3D: Implement bounds above and below the disc:
        height_shape = np.zeros_like(data)
        bounds = [slice(None)] * 3  # i.e.: [:, :, :]
        start = int(np.floor(center[axis] - height/2))
        stop = int(np.ceil(center[axis] + height/2))
        bounds[axis] = slice(start, stop)  # replace with actual bounds along disc symmetry axis
        height_shape[tuple(bounds)] = 1
        data *= height_shape
    return Field(data=data, scale=scale, vector=False)


def create_shape_ellipsoid(dim, center=None, width=None, scale=1):
    """Creates a Field object with the shape of an ellipsoid in arbitrary dimensions.

    Attributes
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple, optional
        The center of the ellipsoid in pixel coordinates.
    width : tuple, optional
        The half axes of the ellipsoid in pixel coordinates.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    if center is None:
        center = tuple([d/2 for d in dim])
    if width is None:
        width = tuple([d/2 for d in dim])
    assert len(dim) == len(center) == len(width), 'Parameters dim, center and width must have the same dimensions!'
    coords = np.indices(dim) + 0.5  # 0.5 to get to pixel/voxel center!
    coords = coords - np.asarray(center)[(slice(None),) + (None,)*len(dim)]  # [:, None, None, None]
    distance = np.sqrt([(coords[i] / (width[i]/2))**2 for i in range(len(dim))])
    data = np.where(distance <= 1, 1, 0)
    return Field(data=data, scale=scale, vector=False)


def create_shape_sphere(dim, center=None, radius=None, scale=1):
    """Creates a Field object with the shape of a sphere in arbitrary dimensions.

    Attributes
    ----------
    dim : tuple
        The dimensions of the grid.
    center : tuple, optional
        The center of the sphere in pixel coordinates.
    width : tuple, optional
        The half axes of the sphere in pixel coordinates.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    if center is None:
        center = tuple([d/2 for d in dim])
    if radius is None:
        radius = np.max(dim) / 4
    assert len(dim) == len(center), 'Parameters dim and center must have the same dimensions!'
    assert radius > 0, 'Radius has to be a positive scalar value!'
    coords = np.indices(dim) + 0.5  # 0.5 to get to pixel/voxel center!
    coords = coords - np.asarray(center)[(slice(None),) + (None,)*len(dim)]  # [:, None, None, None]
    distance = np.sqrt([coords[i]**2 for i in range(len(dim))])
    data = np.where(distance <= radius, 1, 0)
    return Field(data=data, scale=scale, vector=False)


def create_shape_filament(dim, pos=None, axis=0, scale=1):
    """Creates a Field object with the shape of a filament in arbitrary dimension.

    Parameters
    ----------
    dim : tuple
        The dimensions of the grid.
    pos : tuple, optional
        The position of the filament in pixel coordinates. Has to be a tuple of len(dim) - 1 and denotes the
        index positions along all axes aside from the specified `axis` along which the filament is placed.
    axis :  int, optional
        The orientation of the filament axis. Defaults to the first axis.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    assert len(dim) > 1, 'Only usable for multidimensional Fields (at least 2 dimensions)!'
    if pos is None:
        pos = tuple([d/2 for d in dim])
    assert len(dim) == len(pos)-1, 'Position has to specify one less coordinates than dimensions!'
    data = np.zeros(dim)
    index = list(pos)
    index.insert(axis, slice(None))  # [:] for the filament axis! pos everywhere else!
    data[tuple(index)] = 1
    return Field(data=data, scale=scale, vector=False)


def create_shape_voxel(dim, pos=None, scale=1):
    """Creates a Field object with the shape of a single voxel in arbitrary dimension.

    Parameters
    ----------
    dim : tuple
        The dimensions of the grid.
    pos : tuple, optional
        The position of the voxel.
    scale: tuple of float
        Scaling along the dimensions of the underlying data.

    """
    _log.debug('Calling __init__')
    assert len(dim) > 1, 'Only usable for multidimensional Fields (at least 2 dimensions)!'
    if pos is None:
        pos = tuple([d/2 for d in dim])
    assert len(dim) == len(pos), 'Parameters dim and pos must have the same dimensions!'
    data = np.zeros(dim)
    data[pos] = 1
    return Field(data=data, scale=scale, vector=False)
